#! /bin/sh
#
# Configuration script for testbase binary <-> decimal conversion tester.
# Two arguments are required.  The first is the name of the system being
# configured - "sun", for example, for SunOS.  The second is the name of
# the precision to be tested.  For example, under SunOS the second argument
# will be one of sp, dp, qp, or xp, for single, double, quadruple, and
# extended precision.
#
# A third argument, "fortran", may also be given.  If so, it specifies that
# the conversion is to be done in FORTRAN, rather than C.
#

if [ $# -lt 2 ]; then
	echo Too few arguments given to config.sh
	echo Usage: config.sh system precision '[fortran]'
	exit 1
fi

system=$1
precision=$2
fortran=0

if [ $# -eq 3 ]; then
	if [ $3 = "fortran" ]; then
		fortran=1
	else
		echo Third argument, if specified, must be '"fortran"'
		exit 1
	fi
fi

if [ $# -gt 3 ]; then
	echo Too many arguments given to config.sh
	echo Usage: config.sh system precision '[fortran]'
	exit 1
fi

if [ ! -r system/$1-config.h ]; then
	echo No such system type '"'$1'"'
	exit 1
fi

if [ ! -r precision/$1-$2.h ]; then
	echo No such precision '"'$1-$2'"'
	exit 1
fi

rm -f config.h except.c os.c round.c precision.h precision.c fortran.F
rm -f curr-config

cp system/$1-config.h config.h
cp system/$1-except.c except.c
cp system/$1-intr.c intr.c
cp system/$1-os.c os.c
cp system/$1-round.c round.c

cp precision/$1-$2.h precision.h
cp precision/$1-$2.c precision.c

if [ $fortran -eq 1 ]; then
	cp system/$1-fortran.F fortran.F
else
	cat <<! >fortran.F
       subroutine fdummy
       end
!
fi

echo "Current system configuration: " $1 >curr-config
echo "Current precision configuration: " $1-$2 >>curr-config
