/* Testbase routines for generating test cases */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


int
is_to_binary( t )
	enum testtype	t;
{
	/* Return true if the test type indicated by t is one
	 * of decimal-to-binary, as opposed to binary-to-decimal.
	 */
	return t == etob || t == ftob || t == gtob;
}

char *
add_zeroes( str, num_zeroes )
	char	*str;
	int	num_zeroes;
{
	/* Adds num_zeroes worth of '0' characters to the string str,
	 * and returns a pointer to the next position in str after
	 * all of the new zeroes.  This position is also set to NUL,
	 * as a convenience for those cases in which we won't be adding
	 * anything further to the string.
	 */
	
	int	i;

	for (i = 0; i < num_zeroes; i++)
		*str++ = '0';

	*str = 0;

	return str;
}


void
generate_one( t, b, d )
	enum testtype   t;
	FLOAT          *b;
	decimalstring   d;
{
	/* For the test type given by t, produce in b or d (as indicated
	 * by the direction of t) the value 1.0.
	 */

	switch (t) {

	case btoe:
	case btof:
	case btog:
		tenpowerstring((EXPONENT) 0, (EXPONENT) 1, d);
		decimalstring_to_binary(d, b);
		break;

	case etob:
		*d++ = '1';
		*d++ = '.';
		d = add_zeroes( d, (int) cursig - 1 );
		*d++ = 'E';
		*d++ = '-';
		*d++ = '0';
		*d++ = 0;
		break;

	case ftob:
		*d++ = '-';
		*d++ = '1';
		*d++ = '.';
		d = add_zeroes( d, (int) cursig - 1 );
		break;

	case gtob:
		*d++ = '1';
		*d++ = 0;
		break;

	default:
		fatal("bad test type in generate_one()");
	}
}


void
positive_power_of_ten( t, b, d, exponent )
	enum testtype   t;
	FLOAT          *b;
	decimalstring   d;
	EXPONENT	exponent;
{
	/* For the test type given by t, produce in b or d (as indicated
	 * by the direction of t) a positive power (given by "exponent")
	 * of ten.
	 */
	
	char	*pd;

	if ( (t == gtob || t == btog) && exponent >= max_10_f_exponent )
		/* Transform G format to E format. */
		t = (t == gtob ? etob : btoe);

	pd = d;

	*pd++ = random_sign();
	*pd++ = '1';

	switch (t) {
	case etob:
	case btoe:
		*pd++ = '.';
		pd = add_zeroes( pd, (int) cursig - 1 );
		(void) sprintf(pd, "E%+ld", exponent);
		break;

	case ftob:
	case btof:
		pd = add_zeroes( pd, (int) exponent );
		*pd++ = '.';
		pd = add_zeroes( pd, (int) cursig );
		break;

	case gtob:
	case btog:
		pd = add_zeroes( pd, (int) exponent );
		break;

	default:
		fatal("bad test type in positive_power_of_ten()");
	}

	if ( ! is_to_binary( t ) )
		decimalstring_to_binary(d, b);
}


void
negative_power_of_ten( t, b, d, exponent )
	enum testtype   t;
	FLOAT          *b;
	decimalstring   d;
	EXPONENT	exponent;
{
	/* For the test type given by t, produce in b or d (as indicated
	 * by the direction of t) a negative power (given by "exponent")
	 * of ten.
	 */
	
	char	*pd;

	if ( (t == gtob || t == btog) && exponent <= min_10_f_exponent )
		/* Transform G format to E format. */
		t = (t == gtob ? etob : btoe);

	pd = d;
	*pd++ = random_sign();

	switch (t) {
	case etob:
	case btoe:
		*pd++ = '1';
		*pd++ = '.';
		pd = add_zeroes( pd, (int) cursig - 1 );
		(void) sprintf(pd, "E%+ld", exponent);
		break;

	case gtob:
	case btog:
	case ftob:
	case btof:
		*pd++ = '0';
		*pd++ = '.';
		pd = add_zeroes( pd, (int) -(exponent + 1) );
		*pd++ = '1';
		*pd++ = 0;
		break;

	default:
		fatal("bad test type in negative_power_of_ten()");
	}

	if ( ! is_to_binary( t ) )
		decimalstring_to_binary(d, b);
}


void
random_source( t, b, d )
	enum testtype   t;
	FLOAT          *b;
	decimalstring   d;
{
	/* For the test type given by t, produce in b or d (as indicated
	 * by the direction of t) a random source.
	 */

	if ( ! is_to_binary( t ) ) {
		do
			floatrandom(b);
		while (binary_OK(t, b) == 0);

	} else if ( t == etob )
		estringrandom(cursig, d);
	
	else if ( t == ftob )
		fstringrandom(cursig, d);

	else /* t == gtob */
		gstringrandom(cursig, d);
}


void
random_target( t, b, d )
	enum testtype   t;
	FLOAT          *b;
	decimalstring   d;
{
	/* For the test type given by t, produce in b or d (as indicated
	 * by the direction of t) a random target.
	 */

	if ( is_to_binary( t ) ) {	/* etob, ftob, or gtob */
		do
			floatrandom(b);
		while (binary_OK(t, b) == 0);

		binary_to_decimalstring(b, d, cursig, t);

	} else	{
		do {
			switch (t) {
			case btoe:
				estringrandom(cursig, d);
				break;
			case btof:
				fstringrandom(cursig, d);
				break;
			case btog:
				gstringrandom(cursig, d);
				break;

			default:
				fatal("bad test type in random_target()");
			}

			decimalstring_to_binary(d, b);

		} while (binary_OK(t, b) == 0);
	}
}


void
random_target_plus_or_minus( t, b, d, sign )
	enum testtype   t;
	FLOAT          *b;
	decimalstring   d;
	enum fsign	sign;
{
	/* For the test type given by t, produce in b or d (as indicated
	 * by the direction of t) a random target with a positive or negative
	 * half-ulp offset as indicated by "sign".
	 */

	FLOAT           bulp;
	decimalstring   dulp;

	if ( is_to_binary( t ) ) {	/* etob, ftob, or gtob */
		do
			floatrandom(b);
		while (binary_OK(t, b) == 0);

		binary_halfulp(b, FLOAT_ADDR bulp);
		stringsum(b, FLOAT_ADDR bulp, sign, d);

	} else {
		do {
			switch (t) {
			case btoe:
				estringrandom(cursig, d);
				break;
			case btof:
				fstringrandom(cursig, d);
				break;
			case btog:
				gstringrandom(cursig, d);
				break;

			default:
				fatal(
			"bad test type in random_target_plus_or_minus()");
			}

			decimal_plushalfulp(d, sign, dulp);
			decimalstring_to_binary(dulp, b);

		} while (binary_OK(t, b) == 0);
	}
}


void
generate_data(t, b, d)
	enum testtype   t;
	FLOAT          *b;
	decimalstring   d;

{
	/*
	 * Generate the next data item for test t, and put it in *pb if
	 * source is binary, or in d if source is decimal.
	 */

	FLOAT           tb;
	decimalstring   td;
	char           *pd;
	FLOAT          *pb;

	/* Exponents of last powers of ten and two used to generate
	 * test data.
	 */
	static EXPONENT	posten_exponent = 0, negten_exponent = 0;
	static EXPONENT postwo_exponent = 0, negtwo_exponent = 0;

	/*
	 * On d->b, set up pb to point to temporary target tb, pd
	 * to point to actual source d; on b->d, set up pb to point
	 * to actual source b, pd to point to temporary target td.
	 */
	if ( is_to_binary( t ) ) {
		/* If FLOAT is an array type then the following assignment
		 * can generate a warning, since pb's type is pointer-to-array
		 * and not pointer-to-double (or whatever the first element
		 * of tb is).
		 */
		pb = (FLOAT *) (FLOAT_ADDR tb);
		pd = d;

	} else {
		pb = b;
		pd = td;
	}

	/* Move along to the next type of data source. */
	lastdata = (enum datatype) (1 + (int) lastdata);

	if (lastdata == lastdatatype) {
		/* Wrap around to first type of data source. */
		if (skip_powers == 1)
			lastdata = randomsource;
		else
			lastdata = posten;
	}

	switch (lastdata) {
	case startdata:	/* Initialize - shouldn't get here. */
	default:
		fatal("bug in generate_data() - bad lastdata" );

	case onedatum:
		generate_one( t, pb, pd );
		break;

	case posten:
		if (posten_exponent >= max_ten_exponent ||
		    ((t == ftob || t == btof) &&
		     posten_exponent >= max_10_f_exponent)) {
			/* On to the next test. */
			generate_data( t, b, d );
			return;
		}

		posten_exponent++;

		positive_power_of_ten( t, pb, pd, posten_exponent );

		break;

	case negten:
		if (negten_exponent <= min_ten_exponent ||
		    ((t == ftob || t == btof) &&
		     negten_exponent <= min_10_f_exponent)) {
			/* On to the next test. */
			generate_data( t, b, d );
			return;
		}

		negten_exponent--;

		negative_power_of_ten( t, pb, pd, negten_exponent );

		break;

	case postwo:
		if (postwo_exponent >= max_two_exponent ||
		    (t == btof && postwo_exponent >= max_2_f_exponent)) {
			/* On to the next test. */
			generate_data( t, b, d );
			return;
		}

		postwo_exponent++;

		twopowerbinary(postwo_exponent, pb);

		if ( is_to_binary( t ) )
			binary_to_decimalstring(pb, pd, cursig, t);

		break;

	case negtwo:
		if (negtwo_exponent <= min_two_exponent ||
		    ((t == btof || t == ftob) &&
		     negtwo_exponent <= min_2_f_exponent)) {
			/* On to the next test. */
			generate_data( t, b, d );
			return;
		}

		negtwo_exponent--;

		twopowerbinary(negtwo_exponent, pb);

		if ( is_to_binary( t ) )
			binary_to_decimalstring(pb, pd, cursig, t);

		break;

	case randomsource:
		if ( no_stress )
			random_source( t, pb, pd );
		else
			/* Skip this test - stress testing produces plenty
			 * of more-or-less random source values
			 */
			generate_data( t, b, d );
		break;

	case randomtarget:
		random_target( t, pb, pd );
		break;

	case randomtargetplus:
		if (cursig > signoise) {
			/* On to the next test. */
			generate_data( t, b, d );
			return;
		}

		random_target_plus_or_minus( t, pb, pd, positive );
		break;

	case randomtargetminus:
		if (cursig > signoise) {
			/* On to the next test. */
			generate_data( t, b, d );
			return;
		}

		random_target_plus_or_minus( t, pb, pd, negative );
		break;

	case stress:
		if ( no_stress )
			/* On to the next test. */
			generate_data( t, b, d );
		break;
	}
}

void
tenpowerstring(e, sig, s)
	EXPONENT        e, sig;
	decimalstring   s;

{
	/*
	 * Creates an E-format string with sig significant figures and
	 * exponent e.
	 */

	int             i;

	s[0] = random_sign();
	s[1] = '1';
	s[2] = '.';
	for (i = 3; i < (sig + 2); i++)
		s[i] = '0';
	(void) sprintf(&(s[i]), "E%+ld", e);
}

void
twopowerbinary(e, pb)
	EXPONENT        e;
	FLOAT          *pb;

{
	/* Creates *pb = 2**e	more or less.	 */

	unpacked        ub;
	ub.usign = (random_sign() == '+') ? positive : negative;
	ub.uexponent = e;
	ub.uclass = number;
	ub.usignificand.blength = 1;
	ub.usignificand.bsign = ub.usign;
	ub.usignificand.bsignificand[0] = 1;
	unpacked_to_binary(&ub, pb);
}

void
stringsum(pb, pu, sign, s)
	FLOAT          *pb, *pu;
	enum fsign      sign;
	char           *s;

{
	/*
	 * Given FLOAT b and u, compute an E format string representation for
	 * b+u or b-u in s, using cursig significant figures.  sign ==
	 * positive for magnitude add, negative for subtract. This function
	 * is used to generate random string data near halfway points.
	 */

	struct {
		unpacked        u;
		decimalstring   d;
		int             length;
	}               x, y, *px, *py;
	int             i;
	decimalstring   t;
	EXPONENT        shiftcount;
	long            a;
	int             carry;
	char           *ps;

	/* Convert binary_to_decimalstring to E format. */
	binary_to_decimalstring(pb, t, cursig, btoe);
	decimalstring_to_unpacked(t, &(x.u), x.d);
	binary_to_decimalstring(pu, t, cursig, btoe);
	decimalstring_to_unpacked(t, &(y.u), y.d);

	/* Count lengths of d-strings. */
	for (i = 0; x.d[i] != 0; i++)
		;
	x.length = i;
	for (i = 0; y.d[i] != 0; i++)
		;
	y.length = i;
	/* Determine greater and lesser numbers (pg and pl). */
	shiftcount = (x.length + x.u.uexponent) - (y.length + y.u.uexponent);
	if (shiftcount >= 0) {
		px = &x;
		py = &y;
	} else {
		px = &y;
		py = &x;
		shiftcount = -shiftcount;
	}
	/* Right shift lesser to align digits. */
	for (i = py->length; i >= 0; i--)
		py->d[i + shiftcount] = py->d[i];
	for (i = 0; i < shiftcount; i++)
		py->d[i] = '0';
	py->length += shiftcount;
	if (py->length > px->length) {	/* Pad zeros on x. */
		px->u.uexponent -= (py->length - px->length);
		for (i = 0; i < (py->length - px->length); i++)
			px->d[px->length + i] = '0';
		px->d[px->length + i + 1] = 0;
		px->length = py->length;
	}
	/* Add or subtract */
	if (sign == positive) {	/* True add */
		px->d[cursig] += 5;	/* Throw in a round. */
		carry = 0;
		for (i = py->length; i >= 0; i--) {
			a = px->d[i] - '0' + py->d[i] + carry;
			carry = 0;
			while (a > '9') {	/* Because of round, carry
						 * out can be 0, 1, or 2. */
				carry++;
				a -= 10;
			}
			px->d[i] = a;
		}
	} else {		/* True subtract */
		px->d[cursig] += 5;	/* Throw in a round. */
		carry = 0;
		for (i = py->length; i >= 0; i--) {
			a = px->d[i] + '0' - py->d[i] + carry;
			carry = 0;
			while (a < '0') {	/* Because of round, carry
						 * out can be 0, 1, or 2. */
				carry--;
				a += 10;
			}
			while (a > '9') {	/* Because of round, carry
						 * out can be 0, 1, or 2. */
				carry++;
				a -= 10;
			}
			px->d[i] = a;
		}
	}
	if (carry > 0) {	/* carry out */
		for (i = px->length; i >= 0; i--)
			px->d[i + 1] = py->d[i];
		px->d[0] = '0' + carry;
		px->u.uexponent += 1;
	}
	ps = s;
	*ps++ = (px->u.usign == positive) ? '+' : '-';
	(void) strncpy(ps, px->d, (int) cursig);
	ps += cursig;
	(void) sprintf(ps, "E%+ld",
			(px->u.uexponent + (px->length - (int) cursig)));
	if ((s[0] < '+') || (s[0] > 'z'))
		(void) printf("INTERNAL ERROR: stringsum to s >%s< \n", s);
}

void
generate_stress_tests()
{
	unsigned	b1, b2;
	int		d1, d2;
	EXPONENT	e2;
	int		test_near_half_ulp;

	if ( is_to_binary( curtest ) ) {
		b1 = 10;
		b2 = 2;
		d1 = cursig;
		d2 = FLOAT_PRECISION;

		e2 = countregions * bin_stress_exponent_stride;
		e2 %= max_bin_stress_exponent - min_bin_stress_exponent + 1;
		e2 += min_bin_stress_exponent;

	} else {
		EXPONENT	min_exp, max_exp, stride;

		b1 = 2;
		b2 = 10;
		d1 = FLOAT_PRECISION;
		d2 = cursig;

		if ( curtest == btof ) {
			min_exp = min_btof_dec_stress_exponent;
			max_exp = max_btof_dec_stress_exponent;
			stride = btof_dec_stress_exponent_stride;

		} else {
			min_exp = min_dec_stress_exponent;
			max_exp = max_dec_stress_exponent;
			stride = dec_stress_exponent_stride;
		}

		e2 = countregions * stride;
		e2 %= max_exp - min_exp + 1;
		e2 += min_exp;
	}

	if ( curround == round_unbiased || curround == round_biased )
		test_near_half_ulp = 1;
	else
		test_near_half_ulp = 0;

	worst_cases( b1, b2, d1, d2, e2, test_near_half_ulp );
}
