/* Testbase routines for manipulating packed/unpacked values */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"
#include <ctype.h>


void
copy_unpacked(u1, u2)
	unpacked       *u1, *u2;
{
	/* *u2 = *u1	 */

	u2->usign = u1->usign;
	u2->uclass = u1->uclass;
	u2->uexponent = u1->uexponent;
	u2->ubase = u1->ubase;
	bigint_copy(&(u1->usignificand), &(u2->usignificand));
}

void
dump_unpacked(pu)
	unpacked       *pu;
{
	/* Debugging code to display unpacked number in all its glory. */

	EXPONENT        e2, eb;
	int             id;

	eb = pu->uexponent;
	e2 = 0;

	(void) printf("unpacked sign %d class %d significand ",
			pu->usign, pu->uclass);
	if (pu->uclass == number) {
		id = 0;
		while (id < pu->usignificand.blength) {
			if (pu->usignificand.bsignificand[id] != 0)
				(void) printf(
					"\n %c %lX * 2 ** %ld * %u ** %ld ",
					(pu->usign == positive) ? '+' : '-',
					pu->usignificand.bsignificand[id],
					e2, pu->ubase, eb);
			e2 += BIGBASE_BITS;
			id++;
		}
	}
	(void) printf("\n");
	return;
}

void
display_unpacked(pu)
	unpacked       *pu;
{
	/* Display unpacked number in the form digits * base^exp */

	decimalstring	d;

	if ( pu->usign == negative )
		printf( "-" );

	bigint_to_decimalstring( &pu->usignificand, d );

	(void) printf("%s * %u^%ld", d, pu->ubase, pu->uexponent);
}

void
decimalstring_to_unpacked(s, u, d)
	char           *s;
	unpacked       *u;
	decimalstring   d;

{
	/*
	 * Converts decimal string to unpacked form with base=10.  Accepts
	 * SunOS 4.0 notation for infinities and nans.  d gets a string
	 * containing all significant digits encountered corresponding to the
	 * binary significand in u->usignificand.bsignificand.
	 */

	register char  *cp = s;
	char           *cp0 = cp;
	char           *good = cp - 1;	/* End of known good token. */
	char           *pd;
	int             sigfound;
	unsigned        explicitsign = 0, explicitexponent = 0;

	/* Add another digit to unpacked significand. */
#define BUMP(C) \
	{	\
	bigint_times_ten( &(u->usignificand), (BIGBASE_TYPE)(C - '0') );  \
	*pd++ = C;	\
	}

#define UCASE(c) (islower(c) ? toupper(c) : c)

	pd = d;

	u->uclass = number;	/* Defaults. */
	u->usign = positive;	/* Positive. */
	u->ubase = 10;
	u->uexponent = 0;
	u->usignificand.blength = 1;
	u->usignificand.bsignificand[0] = 0;
	u->usignificand.bsign = u->usign;
	sigfound = -1;		/* -1 = no digits found yet. */

	while (*cp == ' ')
		++cp;			/* Skip white space. */

	if (*cp == '+') {
		++cp;

	} else if (*cp == '-') {	/* Negative. */
		u->usign = negative;
		++cp;
	}

	switch (*cp) {
	case '0':
		while (*cp == '0')
			++cp;	 /* Ignore leading zeros. */
		good = cp - 1;
		sigfound = 0;	/* 0 = only zeros found yet. */
		break;

	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		good = cp;
		sigfound = 1;	/* 1 = significant digits found. */
		BUMP(*cp);
		++cp;
		break;

	case 'i':
	case 'I':
		{		/* Try infinity. */
			static char     infstring[] = "INFINITY";
			int             is, iagree;


			++cp;
			is = 1;
			while (UCASE(*cp) == infstring[is] && is <= 8) {
				cp++;
				is++;
			}
			if (UCASE(*cp) == infstring[is])
				iagree = is + 1;
			else
				iagree = is;

			is++;	/* To account for infstring indexing starting
				 * at 0. */

			if (iagree >= 3) {	/* Found syntactically valid
						 * infinity. */
				if (iagree < 8) {	/* INFxxxx */
					good = cp - (is - 3);
				} else {	/* INFINITYxxx */
					good = cp - (is - 8);
				}
				u->uclass = inf;
				sigfound = iagree;
			}
			goto done;
		}

	case 'n':
	case 'N':
		{		/* Try NaN. */
			static char     nanstring[] = "NAN(";
			int             is;

			cp++;
			is = 1;
			while (UCASE(*cp) == nanstring[is]) {
				cp++;
				is++;
			}
			if ((is == 3)) {	/* Found syntactically valid
						 * NaN. */
				good = cp - 1;
				u->uclass = nan;
				sigfound = 1;
			}
			if (is == 4) {	/* Found NaN followed by parenthesis. */
				good = cp - 2;
				u->uclass = nan;
				sigfound = 1;
				while ((*cp != 0) && (*cp != ')')) {
					BUMP(*cp);
					++cp;
				}
				while ((*cp != 0) && (*cp != ')')) {
					/* Pick up rest of string. */
					cp++;
				}
				if (*cp == ')') {
					good = cp;
					cp++;
				}
			}
			goto done;
		}

	case '.':
		++cp;
		/* Try number. */
		goto afterpoint;

	default:
		goto done;
	}

	/* Keep collecting digits till non-digit found. */
	for ( ;; )
		switch (*cp) {
		case '0':
			while (*cp == '0') {	/* Accumulate zero substring. */
				BUMP(*cp);
				cp++;
			}
			good = cp - 1;
			break;

		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			good = cp;
			sigfound = 1;
			BUMP(*cp);
			cp++;
			break;

		case 'E':
		case 'e':
			if (sigfound == -1)
				goto done;
			goto exponent;

		case '.':
			++cp;
			goto afterpoint;

		default:
			goto done;
		}

afterpoint:
	if (sigfound >= 0) {	/* Better accept the point as good, but don't
				 * accept the next character after.  */
		good = cp - 1;	/* Assume cp points past. */
		if (*good != '.')	/* If not, bump good. */
			good++;
	}

	/* Keep collecting digits till non-digit found. */
	for ( ;; )
		switch (*cp) {
		case '0':
			good = cp;

			if (sigfound < 1) {	/* No significant digits
						 * found so far. */
				sigfound = 0;
			} else {
				BUMP(*cp);
			}

			u->uexponent -= 1;	/* All stored. */
			cp++;

			break;

		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			good = cp;

			if (sigfound < 1) {	/* No significant digits
						 * found so far. */
				sigfound = 1;
				BUMP(*cp);
				u->uexponent -= 1;
			} else {/* Significant digits have begun. */
				BUMP(*cp);
				u->uexponent -= 1;	/* All stored. */
			}

			cp++;

			break;

		case 'E':
		case 'e':
			if (sigfound == -1)
				goto done;
			goto exponent;

		default:
			goto done;
		}

exponent:

	if ((*cp != '+') && (*cp != '-')) {
		/* Skip EeDd. */
		cp++;
	}

	if (*cp == '+')
		cp++;
	else if (*cp == '-') {	/* Negative explicit exponent. */
		cp++;
		explicitsign = 1;
	}

	while (('0' <= *cp && *cp <= '9') || *cp == ' ') {
		/* Accumulate explicit exponent. */
		if (*cp == ' ') {	/* Handle blanks. */
			break;
		}
		good = cp;
		if (explicitexponent <= 400000000) {	/* ### */
			explicitexponent = 10 * explicitexponent +
					   *cp - '0';
		}
		++cp;
	}

	if (explicitsign == 1)
		u->uexponent -= explicitexponent;
	else
		u->uexponent += explicitexponent;

done:
	if (sigfound < 1)
		u->uclass = zero;	/* True zero found. */
	if (good >= cp0) {	/* Valid token found. */
	} else {		/* No valid token found. */
		u->usign = positive;
		u->uclass = nan;/* If anyone looks, x will be nan. */
	}

	*pd = 0;		/* Terminate d string. */
}
