/* Routines for manipulating Sun 3 and 4 double precision */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "spdp.c"
#include "ieee-union.h"


void
canonify_random_float(pu)
	floatsizeunion	*pu;
/*ARGSUSED*/
{
	/*
	 * Given a random string of bits in pu, do whatever is necessary
	 * to put them in proper form for representing an arbitrary
	 * random floating-point number.  For example, on some systems
	 * this may include make sure the high bit is set because the
	 * normalization is explicitly present.
	 */
}

void
canonify_bumped_binary( u )
	floatsizeunion	*u;
/*ARGSUSED*/
{
	/* Given a binary float that has been incremented (or decremented)
	 * via direct bumping of the bits in its representation, adjust it
	 * to make sure it remains valid - clear any unused bits that may
	 * have been twiddled, and normalize if in its present form it
	 * is not normal (and it should be normal)
	 */
}

void
binary_to_unpacked(px, pu)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts binary to unpacked form with base=2. Coded for IEEE 754
	 * arithmetic.  Will require recoding for other systems.
	 */

	/*
	 * The following trickery (rather than using frexp) is to insure that
	 * an ulp of zero or a subnormal is the same as an ulp of the
	 * smallest normal.
	 */

	ieeeuniontype   u;
	unsigned        ks, ke, kf1, kf2, kf3;
	BIGBASE_TYPE    b1, b2;

	pu->ubase = 2;
	pu->uclass = number;	/* Default assumption. */

	copyfloat(px, &(u.fsu));

#define COPYFIELD(F) (signbitfirst == 0) ? u.rd.F : u.ld.F

	ks = COPYFIELD(s);
	ke = COPYFIELD(e);
	kf1 = COPYFIELD(f1);
	kf2 = COPYFIELD(f2);
	kf3 = COPYFIELD(f3);

	b2 = (kf1 << (32 - BIGBASE_BITS)) + kf2;
	b1 = kf3;
	if (ke == 0) {		/* subnormal */
		pu->uexponent = 1 - 0x3ff - 52;
	} else {		/* normal */
		b2 |= (0x100000 << (32 - BIGBASE_BITS));
		pu->uexponent = ke - 0x3ff - 52;
	}
	pu->usignificand.blength = 2;
	if (ke == 0x7ff) {	/* Inf or NaN */
		pu->uclass = nan;
		/* inf/nan decision not based on I bit */
		b2 &= ~(0x100000 << (32 - BIGBASE_BITS));
	}

	pu->usign = (ks == 0) ? positive : negative;
	pu->usignificand.bsign = pu->usign;
	pu->usignificand.bsignificand[0] = b1;
	pu->usignificand.bsignificand[1] = b2;
	if ((b1 | b2) == 0) {
		/* Zero and infinity are special cases of zero significand. */
		pu->uclass = (pu->uclass == number) ? zero : inf;
		pu->usignificand.blength = 1;
	}
}

void
unpacked_to_binary(pu, px)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts unpacked to binary form with base=2. Coded for IEEE 754
	 * arithmetic.  Will require recoding for other systems. Only needs
	 * to work for numbers.  Should work exactly for the case where
	 * pu contains exactly FLOAT_PRECISION bits.  Otherwise, doesn't
	 * need to round correctly as the only other uses are to generate
	 * exact powers of two and ulps of random data.
	 */

	/*
	 * The following trickery (rather than using ldexp) is to insure that
	 * an ulp of zero or a subnormal is the same as an ulp of the
	 * smallest normal.
	 */

	ieeeuniontype   u;
	unsigned        ks, ke, kf1, kf2, kf3;
	BIGBASE_TYPE    b1, b2;

	ks = (pu->usign == positive) ? 0 : 1;
	ke = 0;
	kf1 = 0;
	kf2 = 0;
	kf3 = 0;
	if (pu->uclass != zero) {
		b1 = (pu->usignificand.blength < 1) ?
			0 : pu->usignificand.bsignificand[0];
		b2 = (pu->usignificand.blength < 2) ?
			0 : pu->usignificand.bsignificand[1];

		ke = pu->uexponent + 0x3ff + 52;

		/* Caution!  Here we rely on the knowledge that
		 * BIGBASE_BITS is 28.
		 */
		while ((b2 < (1 << 24)) && (ke > 0)) {	/* Normalize */
			b2 = b2 << 1;
			b1 = b1 << 1;
			if (b1 >= BIGBASE) {
				b1 -= BIGBASE;
				b2++;
			};
			ke--;
		}

		if (ke < 1) {		/* Subnormal */
			do {		/* Normalize */
				if ((b2 % 2) == 1) {
					b1 |= BIGBASE;
				}
				b2 = b2 >> 1;
				if ((b1 % 2) == 1) {
					b1 |= 2;	/* Sticky */
				}
				b1 = b1 >> 1;
				ke++;
			} while (ke < 1);
			ke = 0;
		}
		kf1 = b2 >> 4;
		kf2 = b2 & 0xf;
		kf3 = b1;
	}

#define FIELDCOPY(F,kF) \
	if (signbitfirst == 0)	\
		u.rd.F = kF;	\
	else			\
		u.ld.F = kF

	FIELDCOPY(s,ks);
	FIELDCOPY(e,ke);
	FIELDCOPY(f1,kf1);
	FIELDCOPY(f2,kf2);
	FIELDCOPY(f3,kf3);

	copyfloat(&(u.fsu), px);
}
