/* Routines for manipulating Sun 3 and 4 quadruple precision */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"
#include "ieee-union.h"
#include <floatingpoint.h>


void
canonify_random_float(pu)
	floatsizeunion	*pu;

{
}

void
canonify_bumped_binary( u )
	floatsizeunion	*u;
{
}

void
binary_to_decimalstring(x, s, n, t)
	FLOAT          *x;
	decimalstring   s;
	unsigned long   n;
	enum testtype   t;

{
	switch (t) {
	case btoe:
	case etob:
		/* Standard C E format. */
		(void) sprintf(s, "%*.*Le", (int) n + 8, (int) n - 1,
				*(quadruple *) x);
		break;
	case btof:
	case ftob:
		/* Standard C F format. */
		(void) sprintf(s, "%*.*Lf", (int) n + 8, (int) n,
				*(quadruple *) x);
		break;
	case btog:
	case gtob:
		/* Alternate C G format - need to force dot and trailing
		 * zeros in order to get ulps in the right place during
		 * mono checks, to avoid accidental overflow.
		 */
		(void) sprintf(s, "%#*.*Lg", (int) n + 8, (int) n,
				*(quadruple *) x);
		break;

	default:
		fatal( "bad test type in binary_to_decimalstring()" );
	}
}

void
decimalstring_to_binary(s, x)
	char           *s;
	FLOAT          *x;

{
	(void) sscanf(s, "%Lf", x);
}

void
find_floatsize()
{
	floatsize = sizeof(FLOAT) / sizeof(FLOATSIZEUNIT);
}

void
find_signoise()
{
	signoise = 35;
}

void
find_lsuindex()
{
#ifdef i386
	lsuindex = 0;
	signbitfirst = 0;
#else
	lsuindex = floatsize - 1;
	signbitfirst = 1;
#endif
}

void
find_f_exponents()
{
	min_10_f_exponent = -cursig;
	min_2_f_exponent = -cursig * LOG_10_TO_BASE_2;
	max_10_f_exponent = DECIMALSTRING_LENGTH - cursig - 3;
	max_2_f_exponent = max_10_f_exponent * LOG_10_TO_BASE_2;
}

void
find_extremes()
{
	min_ten_exponent = -4950;
	max_ten_exponent = 4930;
	min_two_exponent = -16444;
	max_two_exponent = 16382;
}

void
copyfloat(x, y)
	FLOAT          *x, *y;

{
	int             i;
	floatsizeunion *px, *py;

	px = (floatsizeunion *) x;
	py = (floatsizeunion *) y;
	for (i = 0; i < floatsize; i++)
		py->fsu[i] = px->fsu[i];
}

void
binary_to_unpacked(px, pu)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts binary to unpacked form with base=2. Coded for IEEE 754
	 * arithmetic.  Will require recoding for other systems.
	 */

	/*
	 * The following trickery (rather than using frexp) is to insure that
	 * an ulp of zero or a subnormal is the same as an ulp of the
	 * smallest normal.
	 */

	ieeeuniontype   u;
	unsigned        ks, ke, kf1, kf2, kf3, kf4, kf5, kf6, kf7;
	BIGBASE_TYPE    b1, b2, b3, b4, b5;

	pu->ubase = 2;
	pu->uclass = number;	/* Default assumption. */

	copyfloat(px, FLOAT_ADDR(u.fsu));

#define COPYFIELD(F) (signbitfirst == 0) ? u.rq.F : u.lq.F

	ks = COPYFIELD(s);
	ke = COPYFIELD(e);
	kf1 = COPYFIELD(f1);
	kf2 = COPYFIELD(f2);
	kf3 = COPYFIELD(f3);
	kf4 = COPYFIELD(f4);
	kf5 = COPYFIELD(f5);
	kf6 = COPYFIELD(f6);
	kf7 = COPYFIELD(f7);

	b1 = kf7;
	b2 = (kf5 << (32 - BIGBASE_BITS)) + kf6;
	b3 = (kf3 << (64 - 2 * BIGBASE_BITS)) + kf4;
	b4 = (kf1 << (96 - 3 * BIGBASE_BITS)) + kf2;
	b5 = 0;
	if (ke == 0) {		/* subnormal */
		pu->uexponent = 1 - 0x3fff - 112;
	} else {		/* normal */
		b5 = 1;
		pu->uexponent = ke - 0x3fff - 112;
	}
	pu->usignificand.blength = 5;

	if (ke == 0x7fff) {	/* Inf or NaN */
		pu->uclass = nan;
		b5 = 0;		/* Inf/NaN decision not based on I bit. */
	}

	pu->usign = (ks == 0) ? positive : negative;
	pu->usignificand.bsign = pu->usign;
	pu->usignificand.bsignificand[0] = b1;
	pu->usignificand.bsignificand[1] = b2;
	pu->usignificand.bsignificand[2] = b3;
	pu->usignificand.bsignificand[3] = b4;
	pu->usignificand.bsignificand[4] = b5;
	if ((b1 | b2 | b3 | b4 | b5) == 0) {
		/* Zero and infinity are special cases of zero significand. */
		pu->uclass = (pu->uclass == number) ? zero : inf;
		pu->usignificand.blength = 1;
	}
}

void
unpacked_to_binary(pu, px)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts unpacked to binary form with base=2. Coded for IEEE 754
	 * arithmetic.  Will require recoding for other systems. Only needs
	 * to work for numbers.  Should work exactly for the case where
	 * pu contains exactly FLOAT_PRECISION bits.  Otherwise, doesn't
	 * need to round correctly as the only other uses are to generate
	 * exact powers of two and ulps of random data.
	 */

	/*
	 * The following trickery (rather than using ldexp) is to insure that
	 * an ulp of zero or a subnormal is the same as an ulp of the
	 * smallest normal.
	 */

	ieeeuniontype   u;
	unsigned        ks, ke, kf1, kf2, kf3, kf4, kf5, kf6, kf7;
	BIGBASE_TYPE    b1, b2, b3, b4, b5;

	ks = (pu->usign == positive) ? 0 : 1;
	ke = 0;
	kf1 = 0;
	kf2 = 0;
	kf3 = 0;
	kf4 = 0;
	kf5 = 0;
	kf6 = 0;
	kf7 = 0;
	if (pu->uclass != zero) {
		b1 = (pu->usignificand.blength < 1) ?
			0 : pu->usignificand.bsignificand[0];
		b2 = (pu->usignificand.blength < 2) ?
			0 : pu->usignificand.bsignificand[1];
		b3 = (pu->usignificand.blength < 3) ?
			0 : pu->usignificand.bsignificand[2];
		b4 = (pu->usignificand.blength < 4) ?
			0 : pu->usignificand.bsignificand[3];
		b5 = (pu->usignificand.blength < 5) ?
			0 : pu->usignificand.bsignificand[4];

		ke = pu->uexponent + 0x3fff + 112;

		/* Caution!  Here we rely on the knowledge that
		 * BIGBASE_BITS is 28.
		 */
		while ((b5 == 0) && (ke > 0)) {	/* Normalize */
			b4 = b4 << 1;
			b3 = b3 << 1;
			b2 = b2 << 1;
			b1 = b1 << 1;
			if (b4 >= BIGBASE) {
				b4 -= BIGBASE;
				b5++;
			};
			if (b3 >= BIGBASE) {
				b3 -= BIGBASE;
				b4++;
			};
			if (b2 >= BIGBASE) {
				b2 -= BIGBASE;
				b3++;
			};
			if (b1 >= BIGBASE) {
				b1 -= BIGBASE;
				b2++;
			};
			ke--;
		}

		if (ke < 1) {		/* Subnormal */
			do {		/* Normalize */
				if ((b5 % 2) == 1) {
					b4 |= BIGBASE;
				}
				b5 = b5 >> 1;
				if ((b4 % 2) == 1) {
					b3 |= BIGBASE;
				}
				b4 = b4 >> 1;
				if ((b3 % 2) == 1) {
					b2 |= BIGBASE;
				}
				b3 = b3 >> 1;
				if ((b2 % 2) == 1) {
					b1 |= BIGBASE;
				}
				b2 = b2 >> 1;
				if ((b1 % 2) == 1) {
					b1 |= 2;	/* Sticky */
				}
				b1 = b1 >> 1;
				ke++;
			} while (ke < 1);
			ke = 0;
		}
		kf1 = b4 >> 12;
		kf2 = b4 & 0xfff;
		kf3 = b3 >> 8;
		kf4 = b3 & 0xff;
		kf5 = b2 >> 4;
		kf6 = b2 & 0xf;
		kf7 = b1;
	}

#define FIELDCOPY(F,kF) \
	if (signbitfirst == 0)	\
		u.rq.F = kF;	\
	else			\
		u.lq.F = kF

	FIELDCOPY(s,ks);
	FIELDCOPY(e,ke);
	FIELDCOPY(f1,kf1);
	FIELDCOPY(f2,kf2);
	FIELDCOPY(f3,kf3);
	FIELDCOPY(f4,kf4);
	FIELDCOPY(f5,kf5);
	FIELDCOPY(f6,kf6);
	FIELDCOPY(f7,kf7);

	copyfloat(FLOAT_ADDR(u.fsu), px);
}

void
testbd(x, s, n, t, e)
	FLOAT          *x;
	decimalstring   s;
	unsigned long  *n;
	enum testtype  *t;
	int            *e;

{
	/*
	 * Using routine to be tested, converts binary to decimal string in
	 * format according to parameter n, which might be n significant
	 * digits for E format, n digits after point for F format, n
	 * significant digits for G format
	 */


	clear_exceptions();
	switch (*t) {
	case btoe:
		/* Standard C E format. */
		(void) sprintf(s, "%*.*Le", (int) *n + 8, (int) *n - 1,
				*(quadruple *) x);
		break;

	case btof:
		/* Standard C F format. */
		(void) sprintf(s, "%*.*Lf", (int) *n + 8, (int) *n,
				*(quadruple *) x);
		break;

	case btog:
		/* Standard C G format. */
		(void) sprintf(s, "%*.*Lg", (int) *n + 8, (int) *n,
				*(quadruple *) x);
		break;

	default:
		fatal( "bad test type in testbd()" );
	}

	*e = get_exceptions();
}

void
testdb(s, x, e)
	char           *s;
	FLOAT          *x;
	int            *e;

{
	/*
	 * Using routine to be tested, converts decimal string to binary.
	 */

	clear_exceptions();
	(void) sscanf(s, CONV, x);
	*e = get_exceptions();
}
