/* Routines for manipulating Sun 3 and 4 extended precision */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"
#include "ieee-union.h"
#include <floatingpoint.h>


void
canonify_random_float(pu)
	floatsizeunion	*pu;

{
	pu->fsu[1] |= 0x80000000;	/* Force normalized! */

#ifdef i386
	/* Force i386 exponent field to have same value as m68k. */
	u2 = pu->fsu[2] >> 16;
	pu->fsu[2] = u2 & 0xffff;	/* Clear unused bits of msu. */
#else
	pu->fsu[0] &= 0xffff0000;	/* Clear unused bits of msu. */
#endif
}

void
canonify_bumped_binary( u )
	floatsizeunion	*u;
{
	/* Given a binary float that has been incremented (or decremented)
	 * via direct bumping of the bits in its representation, adjust it
	 * to make sure it remains valid - clear any unused bits that may
	 * have been twiddled, and normalize if in its present form it
	 * is not normal (and it should be normal)
	 */
	int is_not_subnormal;

#ifdef i386
	u->fsu[2] = u->fsu[2] & 0xffff;   /* Clear unused bits. */
	is_not_subnormal = (u->fsu[2] & 0x7fff) != 0;
#else
	u->fsu[0] = u->fsu[0] & 0xffff0000;       /* Clear unused bits. */
	is_not_subnormal = (u->fsu[0] & 0x7fff0000) != 0;
#endif
	if ( is_not_subnormal ) { /* Don't normalize subnormals! */

		if ((u->fsu[1] & 0x80000000) == 0) {     /* Not normalized. */
			u->fsu[1] = u->fsu[1] << 1;
			if ((u->fsu[lsuindex] & 0x80000000) != 0)
				u->fsu[1]++;
			u->fsu[lsuindex] = u->fsu[lsuindex] << 1;
#ifdef i386
			u->fsu[2] -= 0x1;        /* Adjust exponent. */
#else
			u->fsu[0] -= 0x10000;    /* Adjust exponent. */
#endif
		}
	}
}

void
binary_to_decimalstring(x, s, n, t)
	FLOAT          *x;
	decimalstring   s;
	unsigned long   n;
	enum testtype   t;

{
	/* Nonstandard code for types not supported by scanf/sprint	 */

	decimal_mode    dm;
	decimal_record  dr;
	fp_exception_field_type ef;
	char           *pc;
	int             i;

	switch (t) {
	case btoe:
	case etob:
		dm.df = floating_form;
		break;
	case btof:
	case ftob:
		dm.df = fixed_form;
		break;
	case btog:
	case gtob:
		dm.df = fixed_form;
		break;
	default:
		fatal( "bad test type in binary_to_decimalstring()" );
	}
tryagain:
	dm.rd = fp_direction;
	dm.ndigits = n;

	/* Replace this with whatever is appropriate for your non-standard
	 * type.
	 */
	extended_to_decimal(x, &dm, &dr, &ef);

	pc = s;
	if (dr.sign == 1)
		*pc++ = '-';
	switch (dr.fpclass) {
	case fp_zero:
		(void) strcpy(pc, "0");
		break;
	case fp_infinity:
		(void) strcpy(pc, "Infinity");
		break;
	case fp_quiet:
	case fp_signaling:
		(void) strcpy(pc, "NaN");
		break;
	default:
		if (dm.df == fixed_form) {	/* Check fixed form results
						 * for nonzero. */
			for (i = 0; dr.ds[i] == '0'; i++)
				;
			if (dr.ds[i] == 0 || (ef & ~(1 << fp_inexact)) != 0) {
				/*
				 * Try again if significand is all zeros (x
				 * too small) or exception raised (x too
				 * big).
				 */
				if ((t != btog) && (t != gtob))
					(void) printf(
	"ERROR bad operand to binary_to_decimal_string %lX %lX %lX %lX\n",
					(*x)[0], (*x)[1], (*x)[2], (*x)[3]);
				dm.df = floating_form;
				goto tryagain;
			}
		}
		if ((ef & (1 << fp_overflow)) != 0)
			(void) strcpy(pc, "NaN(overflow)");
		else if ((ef & ~(1 << fp_inexact)) != 0) {
			/* If exceptions still raised ... */
			{
				(void) printf(
		"ERROR extended_to_decimal unknown exception %X masked %X\n",
						ef, ef & ~(1 << fp_inexact));
				(void) strcpy(pc, "NaN(unknown-exception)");
			}
		} else {
			(void) sprintf(pc, "%sE%d", dr.ds, dr.exponent);
		}
		break;
	}
}

void
decimalstring_to_binary(s, x)
	char           *s;
	FLOAT          *x;

{
	/* Nonstandard code for types not supported by sscanf        */

	decimal_mode	dm;
	decimal_record	dr;
	fp_exception_field_type ef;
	char		*pc;
	int		nmax;
	enum decimal_string_form	form;
	char		*echar;
	floatsizeunion	*u;

	pc = s;
	nmax = DECIMALSTRING_LENGTH;
	string_to_decimal(&pc, nmax, 0, &dr, &form, &echar);
	dm.rd = fp_direction;
	u = (floatsizeunion *) x;
	decimal_to_extended(x, &dm, &dr, &ef);

	/* Clear unused bits. */
#ifdef i386
	u->fsu[2] = u->fsu[2] & 0xffff;	/* Clear unused bits. */
#else
	u->fsu[0] = u->fsu[0] & 0xffff0000;	/* Clear unused bits. */
#endif
}

void
find_floatsize()
{
	floatsize = sizeof(FLOAT) / sizeof(FLOATSIZEUNIT);
}

void
find_signoise()
{
	signoise = 20;
}

void
find_lsuindex()
{
#ifdef i386
	lsuindex = 0;
	signbitfirst = 0;
#else
	lsuindex = floatsize - 1;
	signbitfirst = 1;
#endif
}

void
find_f_exponents()
{
	min_10_f_exponent = -cursig;
	min_2_f_exponent = -cursig * LOG_10_TO_BASE_2;
	max_10_f_exponent = DECIMALSTRING_LENGTH - cursig - 3;
	max_2_f_exponent = max_10_f_exponent * LOG_10_TO_BASE_2;
}

void
find_extremes()
{
	min_ten_exponent = -4950;
	max_ten_exponent = 4930;
	min_two_exponent = -16444;
	max_two_exponent = 16382;
}

void
copyfloat(x, y)
	FLOAT          *x, *y;

{
	int             i;
	floatsizeunion *px, *py;

	px = (floatsizeunion *) x;
	py = (floatsizeunion *) y;
	for (i = 0; i < floatsize; i++)
		py->fsu[i] = px->fsu[i];
}

void
binary_to_unpacked(px, pu)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts binary to unpacked form with base=2. Coded for IEEE 754
	 * arithmetic.  Will require recoding for other systems.
	 */

	/*
	 * The following trickery (rather than using frexp) is to insure that
	 * an ulp of zero or a subnormal is the same as an ulp of the
	 * smallest normal.
	 */

	ieeeuniontype   u;
	unsigned        ks, ke, kf1, kf2, kf3, kf4;
	BIGBASE_TYPE    b1, b2, b3;

	pu->ubase = 2;
	pu->uclass = number;	/* Default assumption. */

	copyfloat(px, FLOAT_ADDR(u.fsu));

#define COPYFIELD(F) (signbitfirst == 0) ? u.rx.F : u.lx.F

	ks = COPYFIELD(s);
	ke = COPYFIELD(e);
	kf1 = COPYFIELD(f1);
	kf2 = COPYFIELD(f2);
	kf3 = COPYFIELD(f3);
	kf4 = COPYFIELD(f4);

	b1 = kf4;
	b2 = (kf2 << (32 - BIGBASE_BITS)) + kf3;
	b3 = kf1;
	pu->uexponent = ke - 0x3fff - 63;	/* Normal or subnormal. */
	pu->usignificand.blength = 3;

	if (ke == 0x7fff) {	/* Inf or NaN */
		pu->uclass = nan;
		b3 = b3 & ~BIGBASE;	/* Inf/NaN decision not based on I
					 * bit. */
	}

	pu->usign = (ks == 0) ? positive : negative;
	pu->usignificand.bsign = pu->usign;
	pu->usignificand.bsignificand[0] = b1;
	pu->usignificand.bsignificand[1] = b2;
	pu->usignificand.bsignificand[2] = b3;
	if ((b1 | b2 | b3) == 0) {
		/* Zero and infinity are special cases of zero significand. */
		pu->uclass = (pu->uclass == number) ? zero : inf;
		pu->usignificand.blength = 1;
	}
}

void
unpacked_to_binary(pu, px)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts unpacked to binary form with base=2. Coded for IEEE 754
	 * arithmetic.  Will require recoding for other systems. Only needs
	 * to work for numbers.  Should work exactly for the case where
	 * pu contains exactly FLOAT_PRECISION bits.  Otherwise, doesn't
	 * need to round correctly as the only other uses are to generate
	 * exact powers of two and ulps of random data.
	 */

	/*
	 * The following trickery (rather than using ldexp) is to insure that
	 * an ulp of zero or a subnormal is the same as an ulp of the
	 * smallest normal.
	 */

	ieeeuniontype   u;
	unsigned        ks, ke, kf1, kf2, kf3, kf4;
	BIGBASE_TYPE    b1, b2, b3;

	ks = (pu->usign == positive) ? 0 : 1;
	ke = 0;
	kf1 = 0;
	kf2 = 0;
	kf3 = 0;
	kf4 = 0;
	if (pu->uclass != zero) {
		b1 = (pu->usignificand.blength < 1) ?
			0 : pu->usignificand.bsignificand[0];
		b2 = (pu->usignificand.blength < 2) ?
			0 : pu->usignificand.bsignificand[1];
		b3 = (pu->usignificand.blength < 3) ?
			0 : pu->usignificand.bsignificand[2];

		ke = pu->uexponent + 0x3fff + 63;

		/* Caution!  Here we rely on the knowledge that
		 * BIGBASE_BITS is 28.
		 */
		while ((b3 < 0x80) && (ke > 0)) {	/* Normalize */
			b3 = b3 << 1;
			b2 = b2 << 1;
			b1 = b1 << 1;
			if (b1 >= BIGBASE) {
				b1 -= BIGBASE;
				b2++;
			};
			if (b2 >= BIGBASE) {
				b2 -= BIGBASE;
				b3++;
			};
			ke--;
		}

		if (ke < 1) {		/* Subnormal */
			do {		/* Normalize */
				if ((b3 % 2) == 1) {
					b2 |= BIGBASE;
				}
				b3 = b3 >> 1;
				if ((b2 % 2) == 1) {
					b1 |= BIGBASE;
				}
				b2 = b2 >> 1;
				if ((b1 % 2) == 1) {
					b1 |= 2;	/* Sticky */
				}
				b1 = b1 >> 1;
				ke++;
			} while (ke < 1);
			ke = 0;
		}
		kf1 = b3 & 0xff;
		kf2 = b2 >> 4;
		kf3 = b2 & 0xf;
		kf4 = b1;
	}

#define FIELDCOPY(F,kF) \
	if (signbitfirst == 0)	\
		u.rx.F = kF;	\
	else			\
		u.lx.F = kF

	FIELDCOPY(s,ks);
	FIELDCOPY(e,ke);
	FIELDCOPY(f1,kf1);
	FIELDCOPY(f2,kf2);
	FIELDCOPY(f3,kf3);
	FIELDCOPY(f4,kf4);

	if (signbitfirst == 0)
		u.rx.us = 0;
	else
		u.lx.us = 0;	/* Clear unused field. */

	copyfloat(FLOAT_ADDR(u.fsu), px);
}

void
testbd(x, s, n, t, e)
	FLOAT          *x;
	decimalstring   s;
	unsigned long  *n;
	enum testtype  *t;
	int            *e;

{
	/*
	 * Using routine to be tested, converts binary to decimal string in
	 * format according to parameter n, which might be n significant
	 * digits for E format, n digits after point for F format, n
	 * significant digits for G format
	 */

	/* Nonstandard code for types not supported by scanf/sprint	 */

	decimal_mode    dm;
	decimal_record  dr;
	fp_exception_field_type ef;
	char           *pc;
	int             i;

	switch (*t) {
	case btoe:
		dm.df = floating_form;
		break;
	case btof:
		dm.df = fixed_form;
		break;
	case btog:
		dm.df = fixed_form;
		break;
	default:
		fatal( "bad test type in testbd()" );
	}
tryagain:
	dm.rd = fp_direction;
	dm.ndigits = (int) *n;

	extended_to_decimal(x, &dm, &dr, &ef);

	pc = s;
	if (dr.sign == 1)
		*pc++ = '-';
	switch (dr.fpclass) {
	case fp_zero:
		(void) strcpy(pc, "0");
		break;
	case fp_infinity:
		(void) strcpy(pc, "Infinity");
		break;
	case fp_quiet:
	case fp_signaling:
		(void) strcpy(pc, "NaN");
		break;
	default:
		if (dm.df == fixed_form) {	/* Check fixed form results
						 * for nonzero. */
			for (i = 0; dr.ds[i] == '0'; i++)
				;
			if ((dr.ds[i] == 0) ||
			    ((ef & ~(1 << fp_inexact)) != 0)) {
				/*
				 * Try again if significand is all zeros (x
				 * too small) or exception raised (x too
				 * big).
				 */
				if (*t != btog)
					(void) printf(
				"ERROR bad operand to testbd %lX %lX %lX %lX\n",
					(*x)[0], (*x)[1], (*x)[2], (*x)[3]);
				dm.df = floating_form;
				goto tryagain;
			}
		}
		if ((ef & (1 << fp_overflow)) != 0)
			(void) strcpy(pc, "NaN(overflow)");
		else if ((ef & ~(1 << fp_inexact)) != 0) {
			/* If exceptions still raised ... */
				(void) printf(
		"ERROR extended_to_decimal unknown exception %X masked %X\n",
						ef, ef & ~(1 << fp_inexact));
				(void) strcpy(pc, "NaN(unknown-exception)");
		} else {
			(void) sprintf(pc, "%sE%d", dr.ds, dr.exponent);
		}
		break;
	}

	*e = translate_SUNOS_IEEE_exceptions( (int) ef );
}

void
testdb(s, x, e)
	char           *s;
	FLOAT          *x;
	int            *e;

{
	/*
	 * Using routine to be tested, converts decimal string to binary.
	 */

	/* Nonstandard code for types not supported by scanf/sprint        */

	decimal_mode	dm;
	decimal_record	dr;
	fp_exception_field_type	ef;
	char		*pc;
	int		nmax;
	enum decimal_string_form	form;
	char		*echar;
	floatsizeunion	*u;

	pc = s;
	nmax = DECIMALSTRING_LENGTH;
	string_to_decimal(&pc, nmax, 0, &dr, &form, &echar);
	dm.rd = fp_direction;
	u = (floatsizeunion *) x;
	decimal_to_extended(x, &dm, &dr, &ef);
#ifdef i386
	u->fsu[2] = u->fsu[2] & 0xffff;	/* Clear unused bits. */
#else
	u->fsu[0] = u->fsu[0] & 0xffff0000;	/* Clear unused bits. */
#endif
	*e = translate_SUNOS_IEEE_exceptions( (int) ef );
}
