/* Routines for manipulating Vax D format */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#define NO_FIND_EXTREMES
#include "spdp.c"


union float_layout {
	double fp;
	unsigned short words[4];
};


void
canonify_random_float(pu)
	floatsizeunion	*pu;
{
	/*
	 * Given a random string of bits in pu, do whatever is necessary
	 * to put them in proper form for representing an arbitrary
	 * random floating-point number.  For example, on some systems
	 * this may include make sure the high bit is set because the
	 * normalization is explicitly present.
	 */

	unsigned short	e;

	e = (pu->fsu[0] & (0xff << 7)) >> 7;

	if ( e == 0 ) {
		/* Biased exponent is zero.  Avoid reserved operands -
		 * turn the whole float into a zero.
		 */
		pu->f = 0.0;

	} else if ( e == 1 || e == 255 ) {
		/* Avoid exponents too close to underflow or overflow, since
		 * when we add or subtract to them (such as is done by
		 * random_target_plus_or_minus) we risk generating under-
		 * or overflows.
		 */
		if ( e == 1 )
			++e;	/* Move away from underflow. */
		else
			--e;	/* Move away from overflow. */

		pu->fsu[0] &= ~(0xff << 7);	/* Strip out old exponent. */
		pu->fsu[0] |= e << 7;

	}
}

void
canonify_bumped_binary( u )
	floatsizeunion	*u;
{
	/* Given a binary float that has been incremented (or decremented)
	 * via direct bumping of the bits in its representation, adjust it
	 * to make sure it remains valid - clear any unused bits that may
	 * have been twiddled, and normalize if in its present form it
	 * is not normal (and it should be normal)
	 */

	unsigned short	e;

	e = (u->fsu[0] & (0xff << 7)) >> 7;

	if ( e == 0 ) {
		/* Biased exponent is zero.  Avoid reserved operands -
		 * turn the whole float into a zero.
		 */
		u->f = 0.0;

	}
}

void
find_extremes()
{
	/* Find extreme exponents. */

	min_two_exponent = -128;
	max_two_exponent = 126;
	min_ten_exponent = -38;
	max_ten_exponent = 37;
}

void
binary_to_unpacked(px, pu)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts binary to unpacked form with base=2. Coded for Vax D
	 * format.  Will require recoding for other systems.
	 */

	union float_layout	du;
	BIGBASE_TYPE		b1, b2;

	pu->ubase = 2;
	pu->uclass = number;

	copyfloat(px, &du.fp);

	pu->usign = ((du.words[0] & (1 << 15)) == 0) ? positive : negative;
	pu->usignificand.bsign = pu->usign;
	pu->uexponent = ((du.words[0] & (0xff << 7)) >> 7) -
							128 - FLOAT_PRECISION;

	if ( *px == 0.0 ) {
		pu->uclass = zero;
		pu->usignificand.bsignificand[0] = 0;
		pu->usignificand.blength = 1;
		pu->uexponent = 0;

	} else {
		b1 = du.words[3] | (du.words[2] << 16);
		b2 = b1 >> BIGBASE_BITS;
		b1 &= (BIGBASE - 1);
		b2 |= (du.words[1] << 4) | ((du.words[0] & 0x7f) << 20);

		/* Add in the implicit normalization bit. */
		b2 |= (1 << 27);

		pu->usignificand.bsignificand[0] = b1;
		pu->usignificand.bsignificand[1] = b2;
		pu->usignificand.blength = 2;
	}
}

void
unpacked_to_binary(pu, px)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts unpacked to Vax D form.
	 */

	EXPONENT		e;
	union float_layout	du;
	bigint			*b, significand_copy;

	if ( pu->uclass == zero ) {
		*px = 0.0;
		return;
	}

	e = pu->uexponent;
	b = &pu->usignificand;

	if ( b->blength != 2 ||
	     (b->bsignificand[1] & (1 << 27)) == 0 ) {
		/* Normalize significand. */
		EXPONENT	significand_bits, normalization_shift;

		bigint_copy( &pu->usignificand, &significand_copy );
		b = &significand_copy;

		significand_bits = bigint_num_bits( b );
		normalization_shift =
			FLOAT_PRECISION - significand_bits;

		if ( normalization_shift < 0 )
			fatal(
			"bigint too large in unpacked_to_binary()" );

		bigint_times_2_to_power( b, normalization_shift );
		e -= normalization_shift;
	}

	du.words[3] = b->bsignificand[0] & 0xffff;
	du.words[2] = b->bsignificand[0] >> 16;
	du.words[2] |= (b->bsignificand[1] & 0xf) << BIGBASE_BITS;
	du.words[1] = (b->bsignificand[1] >> 4) & 0xffff;
	du.words[0] = b->bsignificand[1] >> 20;

	/* Remove normalization bit. */
	du.words[0] &= 0x7f;

	/* Bias the exponent. */
	e += 128 + FLOAT_PRECISION;

	if ( e <= 0 ) {
		*px = 0.0;
		return;
	}

	du.words[0] |= e << 7;

	if ( pu->usign == negative )
		du.words[0] |= (1 << 15);

	copyfloat(&du.fp, px);
}
