/* Routines for manipulating Vax F format */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#define NO_FIND_EXTREMES
#include "spdp.c"


union float_layout {
	float fp;
	unsigned short words[2];
};


void
canonify_random_float(pu)
	floatsizeunion	*pu;
{
	/*
	 * Given a random string of bits in pu, do whatever is necessary
	 * to put them in proper form for representing an arbitrary
	 * random floating-point number.  For example, on some systems
	 * this may include make sure the high bit is set because the
	 * normalization is explicitly present.
	 */

	unsigned short	e;

	e = (pu->fsu[0] & (0xff << 7)) >> 7;

	if ( e == 0 ) {
		/* Biased exponent is zero.  Avoid reserved operands -
		 * turn the whole float into a zero.
		 */
		pu->f = 0.0;

	} else if ( e == 1 || e == 255 ) {
		/* Avoid exponents too close to underflow or overflow, since
		 * when we add or subtract to them (such as is done by
		 * random_target_plus_or_minus) we risk generating under-
		 * or overflows.
		 */
		if ( e == 1 )
			++e;	/* Move away from underflow. */
		else
			--e;	/* Move away from overflow. */

		pu->fsu[0] &= ~(0xff << 7);	/* Strip out old exponent. */
		pu->fsu[0] |= e << 7;

	}
}

void
canonify_bumped_binary( u )
	floatsizeunion	*u;
{
	/* Given a binary float that has been incremented (or decremented)
	 * via direct bumping of the bits in its representation, adjust it
	 * to make sure it remains valid - clear any unused bits that may
	 * have been twiddled, and normalize if in its present form it
	 * is not normal (and it should be normal)
	 */

	unsigned short	e;

	e = (u->fsu[0] & (0xff << 7)) >> 7;

	if ( e == 0 ) {
		/* Biased exponent is zero.  Avoid reserved operands -
		 * turn the whole float into a zero.
		 */
		u->f = 0.0;

	}
}

void
find_extremes()
{
	/* Find extreme exponents. */

	min_two_exponent = -128;
	max_two_exponent = 126;
	min_ten_exponent = -38;
	max_ten_exponent = 37;
}

void
binary_to_unpacked(px, pu)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts binary to unpacked form with base=2. Coded for Vax F
	 * format.  Will require recoding for other systems.
	 */

	union float_layout	du;
	BIGBASE_TYPE		b1;

	pu->ubase = 2;
	pu->uclass = number;

	copyfloat(px, &du.fp);

	pu->usign = ((du.words[0] & (1 << 15)) == 0) ? positive : negative;
	pu->usignificand.bsign = pu->usign;
	pu->uexponent = ((du.words[0] & (0xff << 7)) >> 7) -
							128 - FLOAT_PRECISION;

	if ( *px == 0.0 ) {
		pu->uclass = zero;
		pu->usignificand.bsignificand[0] = 0;
		pu->usignificand.blength = 1;
		pu->uexponent = 0;

	} else {
		b1 = du.words[1] | ((du.words[0] & 0x7f) << 16);

		/* Add in the implicit normalization bit. */
		b1 |= (1 << 23);

		pu->usignificand.bsignificand[0] = b1;
		pu->usignificand.blength = 1;
	}
}

void
unpacked_to_binary(pu, px)
	FLOAT          *px;
	unpacked       *pu;

{
	/*
	 * Converts unpacked to Vax F form.
	 */

	EXPONENT		e;
	union float_layout	du;
	bigint			*b;

	if ( pu->uclass == zero ) {
		*px = 0.0;
		return;
	}

	e = pu->uexponent;
	b = &pu->usignificand;

	/* Normalize significand. */
	while ( (b->bsignificand[0] & (1 << 23)) == 0 ) {
		b->bsignificand[0] <<= 1;
		--e;
	}

	du.words[1] = b->bsignificand[0] & 0xffff;
	du.words[0] = b->bsignificand[0] >> 16;

	/* Remove normalization bit. */
	du.words[0] &= 0x7f;

	/* Bias the exponent. */
	e += 128 + FLOAT_PRECISION;

	if ( e <= 0 ) {
		*px = 0.0;
		return;
	}

	du.words[0] |= e << 7;

	if ( pu->usign == negative )
		du.words[0] |= (1 << 15);

	copyfloat(&du.fp, px);
}
