/* Function prototypes for all routines used by testbase */

#ifdef __STDC__
#define PROTO(arglist) arglist
#else
#define PROTO(arglist) ()
#endif

bigint	*bigint_create	PROTO( (int d) );
void	bigint_set	PROTO( (bigint *b, int d) );
void	bigint_copy	PROTO( (bigint *b1, bigint *b2) );
void	bigint_display	PROTO( (bigint *b) );
void	bigint_to_decimalstring	PROTO( (bigint *b, decimalstring d) );
void	bigint_grow	PROTO( (bigint *b, int new_size) );
int	bigint_is_zero	PROTO( (bigint *b) );
void	bigint_normalize	PROTO( (bigint *b) );
BIGBASE_TYPE	bigint_normalize_digit	PROTO( (bigint *b, int d) );
int	bigint_compare	PROTO( (bigint *b1, bigint *b2) );
void	bigint_multiply_digits
		PROTO( (bigint *b1, bigint *b2, int d1, int d2, bigint *b3) );
void	bigint_times_half_digit_shifted
		PROTO( (bigint *b1, BIGBASE_TYPE hd, int shift, bigint *b2) );
void	bigint_multiply		PROTO( (bigint *b1, bigint *b2, bigint *b3) );
void	bigint_divide
		PROTO( (bigint *b1, bigint *b2, bigint *b3, bigint *b4) );
void	bigint_times_ten	PROTO( (bigint *b, BIGBASE_TYPE carry) );
void	two_to_power		PROTO( (EXPONENT e, bigint *b) );
bigint	*ten_to_power		PROTO( (EXPONENT e) );
void	bigint_times_2_to_power	PROTO( (bigint *b, EXPONENT e) );
void	bigint_times_power	PROTO( (bigint *b, unsigned base, EXPONENT e) );
void	bigint_add		PROTO( (bigint *b1, bigint *b2, bigint *b3) );
void	bigint_subtract		PROTO( (bigint *b1, bigint *b2, bigint *b3) );
void	bigint_remainder	PROTO( (bigint *b3, bigint *b2, int *pntimes) );
EXPONENT	bigint_num_bits	PROTO( (bigint *b) );


int	is_to_binary	PROTO( (enum testtype t) );
char	*add_zeroes	PROTO( (char *str, int num_zeroes) );
void	generate_one	PROTO( (enum testtype t, FLOAT *b, decimalstring d) );
void	positive_power_of_ten
PROTO( (enum testtype t, FLOAT *b, decimalstring d, EXPONENT exponent) );
void	negative_power_of_ten
PROTO( (enum testtype t, FLOAT *b, decimalstring d, EXPONENT exponent) );
void	random_source	PROTO( (enum testtype t, FLOAT *b, decimalstring d) );
void	random_target	PROTO( (enum testtype t, FLOAT *b, decimalstring d) );
void	random_target_plus_or_minus
	PROTO( (enum testtype t, FLOAT *b, decimalstring d, enum fsign sign) );
void	generate_data	PROTO( (enum testtype t, FLOAT *b, decimalstring d) );
void	tenpowerstring	PROTO( (EXPONENT e, EXPONENT sig, decimalstring s) );
void	twopowerbinary	PROTO( (EXPONENT e, FLOAT *pb) );
void	stringsum
		PROTO( (FLOAT *pb, FLOAT *pu, enum fsign sign, char *s) );
void	generate_stress_tests	PROTO( () );


void	init_for_convergents	PROTO( (bigint *new_a, bigint *new_b) );
void	compute_modulus		PROTO( (bigint *i, bigint *modulus) );
void	next_convergent		PROTO( () );
void	least_i_with_mod_le
			PROTO( (bigint *goal, bigint *i, bigint *ai_mod_b) );
void	least_i_with_mod_ge
			PROTO( (bigint *goal, bigint *i, bigint *ai_mod_b) );
void	closest_modulus_near_target	PROTO( (bigint *target, bigint *s,
	bigint *min_val, bigint *max_val, bigint *distance, int near_zero,
	int (*progress_fcn)()) );


void	copy_unpacked	PROTO( (unpacked *u1, unpacked *u2) );
void	dump_unpacked	PROTO( (unpacked *pu) );
void	display_unpacked	PROTO( (unpacked *pu) );
void	decimalstring_to_unpacked
			PROTO( (char *s, unpacked *u, decimalstring d) );


int	urandom	PROTO( () );
double	random_fraction	PROTO( () );
void	floatrandom	PROTO( (FLOAT *pb) );
void	estringrandom	PROTO( (unsigned long n, decimalstring d) );
void	fstringrandom	PROTO( (unsigned long n, decimalstring d) );
void	gstringrandom	PROTO( (unsigned long n, decimalstring d) );
char	random_sign	PROTO( () );


void	compute_min_and_max_normals
	PROTO( (unsigned base, int d, bigint *min_normal, bigint *max_normal) );
void	compute_power_fraction	PROTO( (unsigned b1, EXPONENT e1,
	unsigned b2, EXPONENT e2, bigint *numerator, bigint *denominator) );
void	find_b1_range	PROTO( (unsigned b1, unsigned b2, int d1, EXPONENT e2,
EXPONENT *e1min_addr, EXPONENT *e1max_addr, bigint *imin, bigint *imax) );
void	worst_cases
PROTO( (unsigned b1, unsigned b2, int d1, int d2, EXPONENT e2, int near_half) );
int	next_toughie
		PROTO( (bigint *s, bigint *distance, bigint *b, int is_best) );
EXPONENT	excess_bits_in_stress_test	PROTO( () );


void	dummy		PROTO( (FLOAT *pf) );
void	workdummy	PROTO( (WORK *pf) );
EXPONENT	gcd	PROTO( (EXPONENT x, EXPONENT y) );
EXPONENT	find_good_stride	PROTO( (EXPONENT range) );
void	interrupt	PROTO( () );
void	fatal		PROTO( (char *s) );
void	usage		PROTO( () );
void	bad_flag	PROTO( (char flag[]) );
void	bad_flag_arg	PROTO( (char flag[]) );
void	parse_args	PROTO( (int argc, char *argv[]) );
void	init		PROTO( (int argc, char *argv[]) );


void	initialize_testrecord	PROTO( (testrecord *t) );
int	testrecord_is_valid	PROTO( (testrecord *t) );
void	copy_testrecord
	PROTO( (testrecord *t1, testrecord *t2, int merge_stress_flag) );
void	merge_stress_excess	PROTO( (testrecord *t1, testrecord *t2) );
void	copy_best_testrecord	PROTO( (testrecord *old_t, testrecord *new_t) );
void	display_testrecord	PROTO( (testrecord *t) );


void	set_normal_rounding	PROTO( () );
int	binary_OK		PROTO( (enum testtype t, FLOAT *pb) );
int	rounding_out_of_bounds
	PROTO( (enum roundingtype t, testrecord *r, int log_error_flag) );
void	summary			PROTO( () );
void	compare			PROTO( (testrecord *t) );
int	mono_direction		PROTO( (unpacked *u1, unpacked *u2) );
void	introduce_an_error	PROTO( (testrecord *t) );
int	one_test		PROTO( (testrecord *t) );
void	btod_monotests		PROTO( () );
void	dtob_monotests		PROTO( () );
void	finished_region		PROTO( () );
void	test_regions		PROTO( () );
void	testloop		PROTO( () );


void	binary_plus_nulps	PROTO( (FLOAT *pb, int n, FLOAT *pbplus) );
int	bump_decimalstring
		PROTO( (int direction, int ulp, decimalstring dstring) );
void	binary_halfulp		PROTO( (FLOAT *pb, FLOAT *pu) );
void	decimal_plushalfulp
		PROTO( (decimalstring d, enum fsign sig, decimalstring du) );
int	binarylsb		PROTO( (FLOAT *pb) );
char	*gstringpulp		PROTO( (decimalstring d) );
void	ulpdiff			PROTO( (testrecord *t) );
int	check_if_extreme_ulp	PROTO( (testrecord *t) );


void	canonify_random_float	PROTO( (floatsizeunion *pu) );
void	canonify_bumped_binary	PROTO( (floatsizeunion *u) );
void	binary_to_decimalstring
	PROTO( (FLOAT *x, decimalstring s, unsigned long n, enum testtype t) );
void	decimalstring_to_binary	PROTO( (char *s, FLOAT *x) );
void	find_floatsize		PROTO( () );
void	find_signoise		PROTO( () );
void	find_lsuindex		PROTO( () );
void	find_f_exponents	PROTO( () );
void	find_extremes		PROTO( () );
void	copyfloat		PROTO( (FLOAT *x, FLOAT *y) );
void	testbd
PROTO((FLOAT *x, decimalstring s, unsigned long *n, enum testtype *t, int *e));
void	testdb			PROTO( (char *s, FLOAT *x, int *e) );


void	binary_to_unpacked	PROTO( (FLOAT *px, unpacked *pu) );
void	unpacked_to_binary	PROTO( (unpacked *pu, FLOAT *px) );


void	clear_exceptions	PROTO( () );
int	get_exceptions		PROTO( () );


void	init_interrupt_handling	PROTO( () );


WORK	seconds			PROTO( () );


void	set_test_rounding	PROTO( (enum roundingtype rounding_mode) );
