/* Testbase routines for generating random inputs */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


int
urandom()
{
	/* Returns random integer from uniform distribution [UMIN, UMAX] */

	/*
	 * Last random number; used to generate next random; type must be
	 * able to hold 0..2,058,650,956 = 0x7AB4854C
	 */
	static unsigned long rgresult = 0;

	rgresult = ((rgresult * 31413) + 1) % 65536;

	if (rgresult > UMAX)
		(void) printf("random result %lu > %d\n", rgresult, UMAX);

	return (int) rgresult;
}

double
random_fraction()
{
	/* Returns a random, uniformly distributed fraction in the
	 * range [0,1).  Note that while uniformly distributed, not
	 * all of the fraction's bits are guaranteed random.
	 */
	return urandom() / (UMAX + 1.0);
}

void
floatrandom(pb)
	FLOAT          *pb;

{
	/*
	 * Returns a random float created from random unsigned. Depends on
	 * 16-bit short unsigned.
	 */

	int             i;
	floatsizeunion *pu;
	int             u1, u2;

	pu = (floatsizeunion *) pb;
	if (lsuindex == 0) {	/* i386 */
		for (i = floatsize - 1; i >= 0; i--) {
			u1 = urandom();
			u2 = urandom();
			pu->fsu[i] = (u1 << 16) | u2;
		}
	} else {		/* m68k/sparc */
		for (i = 0; i < floatsize; i++) {
			u1 = urandom();
			u2 = urandom();
			pu->fsu[i] = (u1 << 16) | u2;
		}
	}
	canonify_random_float(pu);
}

void
estringrandom(n, d)
	unsigned long   n;
	decimalstring   d;

{
	/* Generate a "random" decimal string in E format with n
	 * significant digits.
	 */

	int             t, i;
	char           *pc;
	EXPONENT        e;

	pc = d;
	t = 4.0 * random_fraction();	/* Generate variate from [0,3]. */
	switch (t) {
	case 0:
		*pc++ = '-';
		break;
	case 1:
		*pc++ = ' ';
		break;
	case 2:
		*pc++ = '+';
		break;
	case 3:
		break;
	}

	t = (n + 2.0) * random_fraction();	/* Generate variate from
						 * [0,n+1].	 */

	for (i = 1; i < t; i++) {	/* Generate digits before point. */
		*pc++ = '0' + 10.0 * random_fraction();
	}
	*pc++ = '.';
	for (; i <= n; i++) {	/* Generate digits after point. */
		*pc++ = '0' + 10.0 * random_fraction();
	}
	/* Generate random exponent. */
	e = min_ten_exponent +
		(max_ten_exponent - min_ten_exponent) * random_fraction() - t;

	t = 5.0 * random_fraction();	/* Generate variate from [0,4]. */

	switch (t) {		/* Append various odd trailing characters to
				 * see if scanf knows when to stop. */
	case 0:
	case 1:
		(void) sprintf(pc, "e%ld ?", e);
		break;
	case 2:
		(void) sprintf(pc, "e%+ld+?", e);
		break;
	case 3:
		(void) sprintf(pc, "E%ld-?", e);
		break;
	case 4:
		(void) sprintf(pc, "E%+ld.?", e);
		break;
	}
}

void
fstringrandom(n, d)
	unsigned long   n;
	decimalstring   d;

{
	/* Generate a "random" decimal string in F format with n digits
	 * after the point.
	 */

	int             t, i;
	char           *pc;

	pc = d;
	t = 4.0 * random_fraction();	/* Generate variate from [0,3]. */
	switch (t) {
	case 0:
		*pc++ = '-';
		break;
	case 1:
		*pc++ = ' ';
		break;
	case 2:
		*pc++ = '+';
		break;
	case 3:
		break;
	}

	t = (n + 2.0) * random_fraction();	/* Generate variate from
						 * [0,n+1].	 */

	for (i = 1; i < t; i++) {	/* Generate digits before point. */
		*pc++ = '0' + 10.0 * random_fraction();
	}
	*pc++ = '.';			/* Insert point. */
	for (; i <= n; i++) {	/* Generate digits after point. */
		*pc++ = '0' + 10.0 * random_fraction();
	}

	/* Now generate some more digits after the point to make n in all. */

	for (i = 1; i < t; i++) {
		*pc++ = '0' + (10 * random_fraction());
	}
	*pc = 0;
}

void
gstringrandom(n, d)
	unsigned long   n;
	decimalstring   d;

{
	/* Generate a "random" decimal string in E or F format with n
	 * significant digits.
	 */

	int             t, i;
	char           *pc;
	EXPONENT        e;

	pc = d;
	t = 4.0 * random_fraction();	/* Generate variate from [0,3]. */
	switch (t) {
	case 0:
		*pc++ = '-';
		break;
	case 1:
		*pc++ = ' ';
		break;
	case 2:
		*pc++ = '+';
		break;
	case 3:
		break;
	}

	t = (n + 2.0) * random_fraction();	/* Generate variate from
						 * [0,n+1].	 */

	for (i = 1; i < t; i++) {	/* Generate digits before point. */
		*pc++ = '0' + 10.0 * random_fraction();
	}
	*pc++ = '.';		/* Insert point. */
	for (; i <= n; i++) {	/* Generate digits after point. */
		*pc++ = '0' + 10.0 * random_fraction();
	}
	/* Generate random exponent. */
	e = min_ten_exponent +
		(max_ten_exponent - min_ten_exponent) * random_fraction() - t;

	t = 5.0 * random_fraction();	/* Generate variate from [0,4]. */

	switch (t) {		/* Append various odd trailing characters to
				 * see if scanf knows when to stop. */
	case 0:
		(void) sprintf(pc, "xyz?");
		break;
	case 1:
		(void) sprintf(pc, "e%ld ?", e);
		break;
	case 2:
		(void) sprintf(pc, "e%+ld+?", e);
		break;
	case 3:
		(void) sprintf(pc, "E%ld-?", e);
		break;
	case 4:
		(void) sprintf(pc, "E%+ld.?", e);
		break;
	}
	*pc = 0;
}

char
random_sign()
{
	/* Randomly returns + or -. */

	if (random_fraction() >= 0.5)
		return '+';
	else
		return '-';
}
