/* Testbase definitions */

/* @(#)testbase.h 1.6 91/02/28 SMI	 */

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */

#include "config.h"
#include "precision.h"


#define ABS(x) ((x) >= 0 ? (x) : -(x))
#define MIN(x,y) ((x) <= (y) ? (x) : (y))
#define MAX(x,y) ((x) >= (y) ? (x) : (y))

/* Number of bits in BIGBASE.  Must be an even power of two.  Should be
 * the largest power of 2 such that 10 times the power will still fit
 * into a BIGBASE_TYPE.  (This assumption is used exactly one place in
 * the code: bigbase_times_ten().  This routine, however, has fairly
 * heavy traffic, so it's worth making this concession in overall bigint
 * size.  The value should be even because the following masks assume so
 * (as do the algorithms they're used in, for multiplication and division).
 *
 * NOTE: the various forms of unpacked_to_binary() in precision/sun-*.c were
 * all written with the assumption that BIGBASE is 28.
 */
#define BIGBASE_BITS 28


#define BIGBASE_HALF_BITS (BIGBASE_BITS / 2)

/* Mask for extracting the lower half of a bigint digit. */
#define BIGINT_LOWER_MASK	((1 << BIGBASE_HALF_BITS) - 1)

/* Mask for extracting the upper half of a bigint digit. */
#define BIGINT_UPPER_MASK	(BIGINT_LOWER_MASK << BIGBASE_HALF_BITS)

/* Extracts the lower half of a bigint digit */
#define BIGINT_EXTRACT_LOWER(d)	(d & BIGINT_LOWER_MASK)

/* Extracts the upper half of a bigint digit */
#define BIGINT_EXTRACT_UPPER(d)	((d & BIGINT_UPPER_MASK) >> BIGBASE_HALF_BITS)


#define HALF_RADIX (1 << (BIGBASE_BITS / 2))

/* Extracts the i'th half-digit from b */
#define EXTRACT_HALF_DIGIT(b, i)					\
	((i) % 2 == 0 ?							\
		BIGINT_EXTRACT_LOWER(b->bsignificand[(i) / 2])		\
		: BIGINT_EXTRACT_UPPER(b->bsignificand[(i) / 2]))

/* Sets the i'th half-digit in b to v */
#define SET_HALF_DIGIT(b, i, v)						\
	((i) % 2 == 0 ?							\
		(b->bsignificand[(i) / 2] =				\
			(b->bsignificand[(i) / 2] & BIGINT_UPPER_MASK) | (v))  \
		: (b->bsignificand[(i) / 2] =				\
			(b->bsignificand[(i) / 2] & BIGINT_LOWER_MASK) |  \
			 ((v) << BIGBASE_HALF_BITS)))


#define BIGBASE_TYPE unsigned long	/* Smallest unsigned type that will
					 * hold 2*BIGBASE-1 */

#define BIGBASE (1<<BIGBASE_BITS)	/* Base of bigint integers. */
#define WORK double		/* Working floating-point type for ulp
				 * calculations. */
#define FWORK doubleprecision	/* Fortran working floating-point type. */
#define EXPONENT long		/* Type that can hold exactly all exponents
				 * of interest. */

/* FLOAT_ADDR is an operator that takes the address of a FLOAT type.
 * Note that if FLOAT is an array then the array name itself evaluates
 * to a pointer to the first element of the array, but that's not what
 * we want; we want a pointer to the whole FLOAT.  Hence the use of
 * a cast in that case.
 */
#if FLOAT_IS_ARRAY
#define FLOAT_ADDR (FLOAT *)
#else
#define FLOAT_ADDR &
#endif

#define UMIN 0			/* Minimum returned by urandom. */
#define UMAX 65535		/* Maximum returned by urandom. */

/* These are needed by the FORTRAN conversion routines */
#define BTOE 0			/* Binary to E format. */
#define BTOF 1			/* Binary to F format. */
#define BTOG 2			/* Binary to G format. */
#define ETOB 3			/* E format to binary. */
#define FTOB 4			/* F format to binary. */
#define GTOB 5			/* G format to binary. */

/* Doesn't have to be terribly accurate */
#define LOG_10_TO_BASE_2	3.3219280948873626



/* Following only makes sense to a C compiler! */
#ifndef FDEFS

#include <stdio.h>
#include <string.h>

/**************************************************************************

	TYPE DEFINITIONS

***************************************************************************/

enum fsign {
	positive, negative
};				/* Possible signs of floating-point operands. */

enum fclass {
	zero, number, inf, nan
};				/* Possible classes of floating-point
				 * operands. */

enum roundingtype {		/* Possible rounding rules.	 */
	round_unbiased,		/* IEEE default */
	round_in,		/* IEEE to zero */
	round_up,		/* IEEE to +infinity */
	round_down,		/* IEEE to -infinity */
	round_biased,		/* VAX */
	round_out,		/* IEEE to infinity */
	round_one,		/* Error < 1 ulp. */
	round_two,		/* Error < 2 ulp. */
	lastroundingtype
};

enum verbositytype {
	terse, verbose, tedious
};				/* Verbosity levels.  */


/* Definitions of tests this program can do.	 */
enum testtype {
	btoe = BTOE,		/* Binary to E format. */
	btof = BTOF,		/* Binary to F format. */
	btog = BTOG,		/* Binary to G format. */
	etob = ETOB,		/* E format to binary. */
	ftob = FTOB,		/* F format to binary. */
	gtob = GTOB,		/* G format to binary. */
	lasttesttype
};

/* Possible exception outcomes of tests. */
enum exception_type {
	exception_inexact,
	exception_underflow, 
	exception_overflow,
	exception_invalid
};

/* Possible source and target combinations. */

enum datatype {			/* Possible types of source data for tests of
				 * conversion from source to target format. */
	startdata,		/* Start here. */
	onedatum,		/* Data == 1. */
	posten,			/* Data == 10**posten_exponent */
	negten,			/* Data == 10**-negten_exponent */
	postwo,			/* Data == 2**powtwo_exponent */
	negtwo,			/* Data == 2**-negtwo_exponent */
	randomsource,		/* Data is random in the source format. */
	randomtarget,		/* Data is random in the target format. */
	randomtargetplus,	/* Data is random in the target format + 0.5
				 * ulp */
	randomtargetminus,	/* Data is random in the target format - 0.5
				 * ulp */
	stress,			/* Use optimization to just above/below
				 * 1 or 0.5 ulp */
	lastdatatype
};

typedef union {			/* Standard union for converting between
				 * FLOAT and array of FLOATSIZEUNIT. */
	FLOAT           f;
	FLOATSIZEUNIT   fsu[MAXFLOATSIZE];
}               floatsizeunion;

typedef char    decimalstring[DECIMALSTRING_LENGTH];	/* Standard C decimal
							 * string type. */

/*************************************************************************

	Global types used internally in the testbase program.

***************************************************************************/

/*************************************************************************
	The bigint type consists of a sign and significand.  The significand
is an array of "digits" of base BIGBASE which is chosen, along with
BIGBASE_TYPE, so that all the values 0..BIGBASE-1 can be multiplied by ten
and added to nine without any error.  bsignificand[0] points to the least
significant digit, bsignificand[blength-1] to the most significant.
***************************************************************************/

/* Big integer structure - signed magnitude. */
typedef struct {
	enum fsign      bsign;	/* Sign. */
	short unsigned  blength;	/* Current length of bigint. */
	BIGBASE_TYPE    bsignificand[BIGINTSIZE];	/* Significand. */
}
                bigint;

/* Unpacked binary or decimal floating-point number structure. */
typedef struct {
	enum fsign      usign;	/* Sign. */
	enum fclass     uclass;	/* Class. */
	unsigned        ubase;	/* Base - defined iff fclass == number. */
	EXPONENT        uexponent;	/* Exponent - defined iff fclass ==
					 * number. */
	bigint          usignificand;	/* Significand - defined iff fclass
					 * == number. */
}
                unpacked;	/* The value is sign * significand * base **
				 * exponent */

typedef struct {
	/*
	 * Ulp differences between two unpacked forms are reported as ulps =
	 * 0.5 * halves + extra ; halves are computed exactly as an integral
	 * value, while extra is subject to (inconsequential) roundoff error,
	 * with |extra| <= 0.25 . aulp is |ulps| >= 0, or is -1 if an ulp
	 * difference doesn't exist, as is the case when one of the operands
	 * is infinity or NaN.
	 */
	short int       halves;
	WORK            extra;
	WORK            aulp;
}               unpacked_ulps;

/*
 * Test record of binary and decimal packed and unpacked forms with ulp
 * difference between the unpacked forms.
 */
typedef struct {
	int             index;	/* Region index. */
	int             offset;	/* Mono offset in ulps from center of region. */
	enum datatype   dt;	/* Method for generating data. */
	int e;			/* Exceptions generated in conversion. */
	int correct_e;		/* Exceptions *should* have been generated. */
	unpacked_ulps   ulp;	/* Ulps difference between unpacked binary
				 * and decimal forms. */
	FLOAT           b;	/* Binary number equivalent to d. */
	decimalstring   d;	/* Decimal string equivalent to b. */
	unpacked        ub;	/* Unpacked form of binary number. */
	unpacked        ud;	/* Unpacked form of decimal string. */

	/* The minimum and maximum extra bits in a stress test that led
	 * to this type of error.  If max_excess is < 0 then no stress
	 * tests resulted in this type of error.
	 */
	EXPONENT	min_excess, max_excess;
}               testrecord;


/*****************************************************************************

	GLOBAL VARIABLES

******************************************************************************/

/* System parameters determined dynamically.	 */

/* Size of FLOAT; could use sizeof, but can be done dynamically. */
extern int	floatsize;

/* String index for least significant byte of FLOAT, determined at run time. */
extern int	lsuindex;

/*
 * 1 indicates bit fields are sign-exponent-significand; 0 indicates
 * significand-exponent-sign
 */
extern int             signbitfirst;

/* Test parameters determined at compile or run time.	 */

extern enum verbositytype verbosity;

extern int	sigmin, sigmax;	/* Limits of significance parameter. */

extern int	signoise;	/* The smallest n such that (1en+1) ==
				 	 * (1en-1) == 1en */

extern int	no_exceptions;	/* 0 to test exceptions, 1 to not */

extern int	no_stress;	/* 0 to stress-test, 1 to not */

extern int	no_sign_zero;	/* 0 to test sign of zero; 1 to not (for
				 	 * Fortran) */

extern int	skip_powers;	/* Set to 1 to skip the powers of two and ten
				 * testing and go directly to random data for
				 * debugging purposes. */

/* If non-zero, each time we've done this many conversions, we introduce an
 * artificial error into one to see if we catch it.
 */
extern int	introduce_errors;

/* True if the current conversion has an introduced error, false otherwise. */
extern int	introduced_error;

/* Rounding mode to be tested currently. */
extern enum roundingtype curround;

/* "Normal" rounding for the rest of the program. */
extern enum roundingtype normal_rounding;

extern char	*roundingmodenames[];
extern char	*testnames[];

extern enum testtype   curtest;	/* Test to be performed currently. */

/* Array of "test-enabled" flags. */
extern int	test_enabled[(int) (lasttesttype)];

/* Array of "round-enabled" flags. */
extern int	rounding_enabled[(int) (lastroundingtype)];

/* Time in seconds that a test was started and stopped. */
extern WORK	timestart, timestop;

extern WORK	timelimit;	/* Limit in seconds for a test. */

/* 0 if all data finished, 1 if terminated prematurely */
extern int	timedout;

/* Type for next data to be generated. */
extern enum datatype	lastdata;

/* Largest and smallest powers of ten. */
extern EXPONENT        max_ten_exponent, min_ten_exponent;
/* Largest and smallest powers of two. */
extern EXPONENT        max_two_exponent, min_two_exponent;

extern EXPONENT        max_2_f_exponent, min_2_f_exponent;
extern EXPONENT        max_10_f_exponent, min_10_f_exponent;

/* Largest and smallest decimal powers over which to do stress-testing,
 * if converting binary to decimal.
 */
extern EXPONENT		max_dec_stress_exponent, min_dec_stress_exponent;

/* Same for binary powers, for converting decimal to binary. */
extern EXPONENT		max_bin_stress_exponent, min_bin_stress_exponent;

/* Same for when doing btof conversions; range is restricted since if
 * we try values less than 10^-cursig they'll just get converted to 0. */
extern EXPONENT		max_btof_dec_stress_exponent,
			min_btof_dec_stress_exponent;

/* Amount by which with each new region we move through the stress
 * exponent range.  In other words, for each region we will do stress-
 * testing for the exponent
 *
 *	min_stress_exponent + countregions*stress_exponent_stride % stress_size
 *
 * where stress_size = max_stress_exponent - min_stress_exponent + 1.
 * Thus we want gcd(stress_exponent_stride,stress_size) = 1 so we'll be sure
 * to step through all of the stress exponents.
 */
extern EXPONENT	dec_stress_exponent_stride, bin_stress_exponent_stride;
extern EXPONENT	btof_dec_stress_exponent_stride;

/*
 * Powers corresponding to maximum and minimum magnitudes that can be
 * represented in F format. If the length of the F format string is
 * constrained to s and the F significance parameter is n (so that n digits
 * are to appear after the point) then testable magnitudes must satisfy
 * 
 *		2**e >= 1 * 10 ** -n, else numbers will be converted to zeros
 *		2**e <= 10**(s-n-3) - 1, else numbers would overflow strings
 *					 if converted correctly
 */

extern EXPONENT	unpacked_one_exponent;	/* exponent of 1 in unpacked struct. */

/* Current significance parameter - corresponding to following error
 * information.
 */
extern unsigned long	cursig;

extern unsigned long   nregions;	/* Number of regions per test. */
/* Number of ulps to look each way during monotonicity tests. */
extern int	nmono;
extern WORK	ulpsbound;	/* For small cursig, too many mono steps
				 * could cause stepping over the overflow
				 * threshold when converting decimal to
				 * binary.  ulpsbound is a bound on the
				 * number of mono steps that can be safely
				 * tolerated. */

/* Test records for center of monotonicity region and monotonicity test
 * points.
 */
extern testrecord	center, mo[2];

/*
 * Cumulative counts of regions, class failures, sign failures, monotonicity
 * failures, rounding errors, conversion tests, introduced errors, introduced
 * errors caught, and introduced errors missed.
 */
extern unsigned long	countregions, countclass, countexcep, countsign,
			countmono, countround, countconversions, countintro,
			count_intro_caught, count_intro_missed;

/* Total number of conversions and regions over all tests performed. */
extern unsigned long	total_conversions, total_regions;

extern testrecord	excepexample;	/* Last example of exception error. */
extern testrecord	classexample;	/* Last example of class error. */
extern testrecord	signexample;	/* Last example of sign error. */
/* Last example of monotonicity error. */
extern testrecord	monoexample1, monoexample2;
extern testrecord	roundexample;	/* Last example of rounding error. */
extern testrecord	whopperexample;	/* Best example w/ > 2 ULPs error. */

/*
 * Examples of largest ulp errors encountered for positive and negative
 * operands.
 */
extern testrecord	ulpminpos, ulpmaxpos, ulpminneg, ulpmaxneg;

/* Convenient bigint constants. */
extern bigint		big_zero, big_one;


#include "protos.h"


#endif	/* ! FDEFS */
