/* Testbase routines for recording the results of tests */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


void
initialize_testrecord(t)
	testrecord     *t;
{
	(void) strcpy(t->d, "uninitialized");
	t->index -= 1;
	t->offset = 123456789;
	t->dt = lastdatatype;
	t->ulp.halves = 0;
	t->ulp.extra = 0;
	t->ulp.aulp = -1;
	t->ub.usign = (enum fsign) 0;
	t->ub.uclass = (enum fclass) 0;
	t->ub.uexponent = 0;
	t->ub.ubase = 0;
	t->ub.usignificand.bsign = (enum fsign) 0;
	t->ub.usignificand.blength = 0;
	t->ud.usign = (enum fsign) 0;
	t->ud.uclass = (enum fclass) 0;
	t->ud.uexponent = 0;
	t->ud.ubase = 0;
	t->ud.usignificand.bsign = (enum fsign) 0;
	t->ud.usignificand.blength = 0;

	t->max_excess = -1;	/* < 0 flags that no stress tests encountered */
}

int
testrecord_is_valid( t )
	testrecord	*t;
{
	/* Returns true if the given test record has been set, false
	 * if it's still in the initialize_testrecord form.
	 */
	return t->ud.ubase != 0;
}

void
merge_stress_excess( t1, t2 )
	testrecord	*t1, *t2;
{
	/* Takes the min and max excess stress bits from test record t1
	 * and merges them into those in t2.  If t1 doesn't have any
	 * stress excess values in it and we just did a stress test,
	 * we add the test's excess bit values into t1 before the merge.
	 */

	if ( t1->max_excess < 0 )
		/* No excess values in t1.  We're done. */
		return;

	else if ( t2->max_excess < 0 ) {
		/* No excess values in t2.  Just take t1's. */
		t2->min_excess = t1->min_excess;
		t2->max_excess = t1->max_excess;

	} else {
		t2->min_excess = MIN(t1->min_excess, t2->min_excess);
		t2->max_excess = MAX(t1->max_excess, t2->max_excess);
	}
}

void
copy_testrecord(t1, t2, merge_stress_flag)
	testrecord	*t1, *t2;
	int		merge_stress_flag;
{
	/* *t2 = *t1.  If merge_stress_flag is true then we merge the
	 * excess stress bit counts from t1 with t2's; otherwise we
	 * just take whatever values t1 already has.  If t1 has no
	 * values and we just did a stress test, we add the test's
	 * excess bit values into t1 before the copy.
	 */

	t2->index = t1->index;
	t2->offset = t1->offset;
	t2->dt = t1->dt;
	copyfloat(FLOAT_ADDR (t1->b), FLOAT_ADDR (t2->b));
	(void) strcpy(t2->d, t1->d);
	t2->ulp.halves = t1->ulp.halves;
	t2->ulp.extra = t1->ulp.extra;
	t2->ulp.aulp = t1->ulp.aulp;
	t2->e = t1->e;
	t2->correct_e = t1->correct_e;
	copy_unpacked(&(t1->ub), &(t2->ub));
	copy_unpacked(&(t1->ud), &(t2->ud));

	if ( t1->max_excess < 0 && lastdata == stress )
		t1->min_excess = t1->max_excess = excess_bits_in_stress_test();

	if ( merge_stress_flag )
		merge_stress_excess(t1, t2);
	else {
		t2->min_excess = t1->min_excess;
		t2->max_excess = t1->max_excess;
	}
}

void
copy_best_testrecord( old_t, new_t )
	testrecord	*old_t, *new_t;
{
	/* Given a choice between two test records, pick the one with
	 * the least-magnitude exponent.  These are the tests that are
	 * easiest to verify.  The best testrecord is stored in old_t.
	 */

	if ( ! testrecord_is_valid( old_t ) ||
	     ABS(old_t->ub.uexponent) > ABS(new_t->ub.uexponent) )
		copy_testrecord(new_t, old_t, 1);

	else {
		/* Even though new_t isn't "better", we still want to
		 * merge in its stress excess contribution.
		 */
		if ( new_t->max_excess < 0 && lastdata == stress )
			new_t->min_excess = new_t->max_excess =
				excess_bits_in_stress_test();
		merge_stress_excess(new_t, old_t);
	}
}

void
describe_exceptions( e )
	int		e;
{
	if ( ! e )
		printf( " none" );

	else {
		if ( e & (1 << exception_inexact) )
			printf( " inexact" );
		if ( e & (1 << exception_underflow) )
			printf( " underflow" );
		if ( e & (1 << exception_overflow) )
			printf( " overflow" );
		if ( e & (1 << exception_invalid) )
			printf( " invalid" );
	}
}

void
display_testrecord(t)
	testrecord     *t;
{	/* Displays testrecord's binary, decimal, and ulps. */

	(void) printf("region %4d: ", t->index);
	switch (t->dt) {
	case onedatum:
		(void) printf("%12s", "One");
		break;
	case posten:
		(void) printf("%12s", "10**n");
		break;
	case negten:
		(void) printf("%12s", "10**-n");
		break;
	case postwo:
		(void) printf("%12s", "2**n");
		break;
	case negtwo:
		(void) printf("%12s", "2**-n");
		break;
	case randomsource:
		(void) printf("%12s", "source");
		break;
	case randomtarget:
		(void) printf("%12s", "target");
		break;
	case randomtargetplus:
		(void) printf("%12s", "target+0.5ut");
		break;
	case randomtargetminus:
		(void) printf("%12s", "target-0.5ut");
		break;
	case stress:
		(void) printf("%12s", "stress");
		break;
	default:
		fatal( "bad datatype in display_testrecord()" );
		break;
	}

	if ( t->offset )
		(void) printf("%+d", t->offset);

	if ( is_to_binary( curtest ) ) {
		(void) printf("\n\tdecimal representation: %*s ",
			(int) (cursig + 8), t->d);

		(void) printf("\n\tbinary representation: ");
		display_unpacked( &t->ub );

	} else {
		(void) printf("\n\tbinary representation: ");
		display_unpacked( &t->ub );

		(void) printf("\n\tdecimal representation: %*s ",
			(int) (cursig + 8), t->d);
	}

	(void) printf("\n");
	if (t->ulp.aulp >= 0) {
		printf( "\tULPS: " );

		if ((t->ulp.halves == 0) && (t->ulp.extra == 0))
			(void) printf("0\n");
		else {
			if (t->ulp.halves == 0) {
				(void) printf("%+6.1e", (double) t->ulp.extra);
			} else {
				if (t->ulp.extra == 0) {
					(void) printf("%+5.1f",
						(double) 0.5 * t->ulp.halves);
				} else {
					if (t->ulp.halves > 0) {
						(void) printf("+(%3.1f",
						  (double) 0.5 * t->ulp.halves);
						(void) printf("%+6.1e)",
							(double) t->ulp.extra);
					} else {
						(void) printf("-(%3.1f",
						 (double) -0.5 * t->ulp.halves);
						(void) printf("%+6.1e)",
							-(double) t->ulp.extra);
					}
				}
			}
			(void) printf("\n");
		}
	}

	if ( t->e != t->correct_e ) {
		(void) printf("\texceptions reported:");
		describe_exceptions( t->e );
		(void) printf("; correct is:");
		describe_exceptions( t->correct_e );
		(void) printf("\n");
	}

	if ( t->max_excess >= 0 ) {
		if ( t->min_excess == t->max_excess ) {
			(void) printf(
			"\tthis value contained %ld excess stress bits\n",
					t->max_excess );

		} else {
			(void) printf(
		"\tmin excess stress bits causing this type of error: %ld\n",
					t->min_excess );
			(void) printf(
		"\tmax excess stress bits causing this type of error: %ld\n",
					t->max_excess );
		}
	}
}
