/* Testbase routines for performing tests */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


void
set_normal_rounding()
{
	/* Set the rounding mode to be used the rest of the time. */

	if (curround == normal_rounding)
		return;

	set_test_rounding( normal_rounding );
}

int
binary_OK(t, pb)
	enum testtype   t;
	FLOAT          *pb;

{
	/*
	 * Tests *pb against various criteria designed to insure that it's
	 * acceptable data for the test in mind.
	 */
	/* Returns 1 for OK, 0 for not OK.		 */

	unpacked        ub;

	binary_to_unpacked(pb, &ub);

	if (ub.uclass != number)
		return 0;

	if (ub.uexponent > (max_two_exponent + unpacked_one_exponent))
		return 0;

	if (ub.uexponent < (min_two_exponent + unpacked_one_exponent))
		return 0;

	if ((t == btof) || (t == ftob) || (t == btog) || (t == gtob)) {
		if (ub.uexponent > (max_2_f_exponent + unpacked_one_exponent))
			return 0;

		if (ub.uexponent < (min_2_f_exponent + unpacked_one_exponent))
			return 0;
	}

	return (1);
}

int
rounding_out_of_bounds(r, t, log_error_flag)
	testrecord     *t;
	enum roundingtype r;
	int		log_error_flag;

{
	/*
	 * Returns 0 if the ulp info in t->ulp is consistent with the sign in
	 * t->ub.usign and the test direction r; 1 if inconsistent.  If
	 * log_error_flag is true and the rounding is inconsistent then
	 * the test record will be remembered in roundexample.
	 */

	unpacked_ulps  *u;

	u = &(t->ulp);

	if (u->aulp < 0)
		goto good;	/* Consistent if not defined. */

	if (ABS(u->extra) > 0.25) {
		/* Error */
	} else switch (r) {
	case round_unbiased:
		if (ABS(u->halves) > 1)
			break;
		if (u->halves == 0)
			goto good;
		if ((u->halves == 1) && (u->extra > 0))
			break;
		if ((u->halves == -1) && (u->extra < 0))
			break;
		if ((ABS(u->halves) == 1) && (u->extra == 0)) {
			/* Ambiguous case - is target even? */
			char           *pu;

			if (is_to_binary( curtest )) {
				if (binarylsb(FLOAT_ADDR (t->b)) != 0)
					/* rounded halfway case to odd! */
					break;
			} else {
				/* id points to the last significand digit
				 * of the string
				 */
				pu = gstringpulp(t->d);
				if (((*pu) % 2) == 1)
					/* rounded halfway case to odd! */
					break;
			}
		}
		goto good;

	case round_biased:
		if (ABS(u->halves) > 1)
			break;
		if (u->halves == 0)
			goto good;
		if (t->ub.usign == negative) {	/* Round biased, negative. */
			if ((u->halves == 1) && (u->extra > 0))
				break;
			if ((u->halves == -1) && (u->extra <= 0))
				break;
		} else {	/* Round biased, positive. */
			if ((u->halves == 1) && (u->extra > 0))
				break;
			if ((u->halves == -1) && (u->extra <= 0))
				break;
		}
		goto good;

	case round_in:
		/* Round to zero: -1 <  ulps <= 0 */
		if (ABS(u->halves) > 2)
			break;
		if (u->halves > 0)
			break;
		if ((u->halves == 0) && (u->extra > 0))
			break;
		if ((u->halves == -2) && (u->extra <= 0))
			break;
		goto good;

	case round_out:
		/* Round to inf: 0 <= ulps < 1.0 */
		if (ABS(u->halves) > 2)
			break;
		if (u->halves < 0)
			break;
		if ((u->halves == 0) && (u->extra < 0))
			break;
		if ((u->halves == 2) && (u->extra >= 0))
			break;
		goto good;

	case round_up:
		if (ABS(u->halves) > 2)
			break;
		if (t->ub.usign == negative) {
			/* Round to +inf, negative number: -1 < ulps <= 0 */
			if (u->halves > 0)
				break;
			if ((u->halves == 0) && (u->extra > 0))
				break;
			if ((u->halves == -2) && (u->extra <= 0))
				break;
			goto good;
		} else {
			/* Round to +inf, positive number: 0 <= ulps < 1 */
			if (u->halves < 0)
				break;
			if ((u->halves == 0) && (u->extra < 0))
				break;
			if ((u->halves == 2) && (u->extra >= 0))
				break;
			goto good;
		}

	case round_down:
		if (ABS(u->halves) > 2)
			break;
		if (t->ub.usign == positive) {
			/* Round to -inf, positive number: -1 < ulps <= 0 */
			if (u->halves > 0)
				break;
			if ((u->halves == 0) && (u->extra > 0))
				break;
			if ((u->halves == -2) && (u->extra <= 0))
				break;
			goto good;
		} else {
			/* Round to -inf, negative number: 0 <= ulps < 1 */
			if (u->halves < 0)
				break;
			if ((u->halves == 0) && (u->extra < 0))
				break;
			if ((u->halves == 2) && (u->extra >= 0))
				break;
			goto good;
		}

	case round_one:
		if (ABS(u->halves) > 2)
			break;
		if (ABS(u->halves) < 2)
			goto good;
		if (u->extra == 0)
			break;
		if ((u->extra > 0) == (u->halves > 0))
			break;
		else
			goto good;

	case round_two:
		if (ABS(u->halves) > 4)
			break;
		if (ABS(u->halves) < 4)
			goto good;
		if (u->extra == 0)
			break;
		if ((u->extra > 0) == (u->halves > 0))
			break;
		else
			goto good;

	default:
		fatal("bad rounding type in rounding_out_of_bounds()");
	}

	/* If we got this far then there was a rounding error. */

	if ( introduced_error ) {
		++count_intro_caught;
		return 0;
	}

	if ( log_error_flag ) {
		++countround;
		copy_best_testrecord( &roundexample, t );
	}

	return 1;

good:
	/* No error found - should we have found one? */
	if ( introduced_error ) {
		++count_intro_missed;
	}
	return 0;
}

void
summary()
{				/* Print summary of errors encountered for
				 * this type of conversion.	 */
	int             outofbounds;
	int		maxposout, maxnegout, minposout, minnegout;

	/*
	 * For /valid purposes, put timing information on standard
	 * error rather than standard out.
	 */
	(void) printf( 
	"\nSummary for %s tests with %lu significance and %s rounding:\n",
			testnames[(int) curtest], cursig,
			roundingmodenames[(int) curround] );

	if ( timelimit == 0.0 )
		(void) printf( "\t%lu conversions in %lu completed regions.\n",
			countconversions, countregions );
	else
		(void) printf( 
	"\t%lu conversions in completed %lu regions took %ld seconds.\n",
				countconversions, countregions,
				(long) (timestop - timestart) );

	if (countclass > 0) {
		(void) printf("%lu CLASS ERRORS including:\n", countclass);
		(void) printf("CLASS ");
		display_testrecord(&classexample);
	}

	if (countsign > 0) {
		(void) printf("%lu SIGN ERRORS including:\n", countsign);
		(void) printf("SIGN ");
		display_testrecord(&signexample);
	}

	if (countexcep > 0) {
		(void) printf("%lu EXCEPTION ERRORS including:\n", countexcep);
		(void) printf("EXCEPTION ");
		display_testrecord(&excepexample);
	}

	if (countround > 0) {
		(void) printf("%lu ROUNDING ERRORS including:\n", countround);
		(void) printf("ROUNDING ");
		display_testrecord(&roundexample);
	}

	if (countmono > 0) {
		(void) printf("%lu MONOTONICITY ERRORS including:\n",
				countmono);
		(void) printf("MONO1 ");
		display_testrecord(&monoexample1);
		(void) printf("MONO2 ");
		display_testrecord(&monoexample2);
	}


	introduced_error = 0;
	maxposout = rounding_out_of_bounds(curround, &ulpmaxpos, 0);
	minposout = rounding_out_of_bounds(curround, &ulpminpos, 0);
	maxnegout = rounding_out_of_bounds(curround, &ulpmaxneg, 0);
	minnegout = rounding_out_of_bounds(curround, &ulpminneg, 0);

	outofbounds = maxposout || minposout || maxnegout || minnegout;
	if (outofbounds) {
		(void) printf("LARGE ULP ERRORS found including:\n");

		if ( minposout ) {
			(void) printf("min+ ");
			display_testrecord(&ulpminpos);
		}

		if ( maxposout ) {
			(void) printf("max+ ");
			display_testrecord(&ulpmaxpos);
		}

		if ( minnegout ) {
			(void) printf("min- ");
			display_testrecord(&ulpminneg);
		}

		if ( maxnegout ) {
			(void) printf("max- ");
			display_testrecord(&ulpmaxneg);
		}

	}

	if ( testrecord_is_valid( &whopperexample ) ) {
		(void) printf( "Best large example:\n     " );
		display_testrecord(&whopperexample);
	}

	if ( countintro > 0 ) {
		(void) printf(
			"%lu errors introduced, of which %lu were detected.\n",
			countintro, count_intro_caught );

		if ( countintro != count_intro_caught )
			(void) printf(
				"NOT ALL INTRODUCED ERRORS WERE DETECTED.\n" );

		if ( count_intro_caught + count_intro_missed != countintro )
			(void) printf(
		"Inconsistencies in introduced error bookkeeping detected.\n" );
	}

	if ( ! outofbounds && countclass == 0 && countsign == 0 &&
	     countmono == 0 && countexcep == 0 && countround == 0 )
		(void) printf("No errors detected.\n");

	(void) printf("\n");
	(void) fflush(stdout);
}

void
compare(t)
	testrecord     *t;

{

	/*
	 * Compare binary and decimal, counting errors, printing messages,
	 * and computing ulps.
	 */

	int             extremeulp, correct;

	correct = 1 << exception_inexact;

	/* Ulp difference not defined unless both numbers. */
	t->ulp.aulp = -1;
	t->correct_e = t->e;	/* Assume okay till we know better. */

	if (t->ub.usign != t->ud.usign) {	/* SIGN ERROR */
		if (no_sign_zero == 1 &&
		    t->ub.uclass == zero && t->ud.uclass == zero) {
			/* Don't test sign of zero for Fortran. */
		} else if ( introduced_error )
			return;
		else {
			countsign++;
			copy_best_testrecord( &signexample, t );
			if ((int) verbosity >= (int) verbose) {
				(void) printf("SIGN ERROR: %d vs %d\n",
						t->ub.usign, t->ud.usign);
				display_testrecord(t);
			}
			return;
		}
	}

	if (is_to_binary( curtest )) {
		if (t->ud.uclass == number) {
			if (t->ub.uclass == zero) {
				/* number -> zero on underflow; check ulps */
				correct |= (1<<exception_underflow);
				goto checkulp;
			} else if (t->ub.uclass == inf) {
				/* number -> inf on overflow; ignore ulp */
				correct |= (1<<exception_overflow);
				goto checkexceptions;	
			}
		}
	}
	if (t->ub.uclass != t->ud.uclass) {
		if ( introduced_error )
			return;
		countclass++;
		copy_best_testrecord( &classexample, t );
		if ((int) verbosity >= (int) verbose) {
			(void) printf("CLASS ERROR: %d vs %d\n",
					t->ub.uclass, t->ud.uclass);
			display_testrecord(t);
		}
		return;
	}
	/* same class, check ulps if both numbers */
	if (t->ub.uclass != number) {
		/* Both binary and decimal representations were non-numbers. */
		correct = 0;
		goto checkexceptions;
	}

checkulp:
	ulpdiff(t);

	if (t->ulp.halves == 0 && t->ulp.extra == 0)
		correct = 0;	/* Conversion should have been exact */

	extremeulp = check_if_extreme_ulp( t );

	if (verbosity == tedious || (verbosity == verbose && extremeulp == 1))
		display_testrecord(t);

checkexceptions:
	/* Check correct exceptions here. */
	if (no_exceptions == 0) {
		if (correct != t->e) { 
			/* If only difference is overflow and underflow
			 * and inexact is signaled then OK.
			 */
			if ((correct ^ t->e) == 1<<exception_underflow ||
			    (correct ^ t->e) == 1<<exception_overflow)
				return;

			countexcep++;

			t->correct_e = correct;

			copy_best_testrecord( &excepexample, t );

			if ((int) verbosity >= (int) verbose) {
				(void) printf(
					"EXCEPTION ERROR: expected %X vs %X\n",
						(unsigned) correct,
						(unsigned) t->e );
				display_testrecord(t);
			}
		}
	}
}

int
mono_direction(u1, u2)
	unpacked       *u1, *u2;

{
	/*
	 * Given two unpacked numbers with the same base, return: -1 if u1 <
	 * u2, 0 if u1 = u2, 1 if u1 > u2, for monotonicity checking.  Signs
	 * are ignored.
	 */

	int             id;
	long            a, borrow;
	int             diff;

	if (u1->uexponent > u2->uexponent) {	/* Scale u1 significand up. */
		bigint_times_power(&(u1->usignificand),
			u1->ubase, u1->uexponent - u2->uexponent);
		u1->uexponent = u2->uexponent;
	}
	if (u1->uexponent < u2->uexponent) {	/* Scale u2 significand up. */
		bigint_times_power(&(u2->usignificand),
			u2->ubase, u2->uexponent - u1->uexponent);
		u2->uexponent = u1->uexponent;
	}
	id = 0;
	diff = 0;		/* Assume equal until proven otherwise. */
	borrow = 0;
	while (id < u1->usignificand.blength && id < u2->usignificand.blength) {
		a = u1->usignificand.bsignificand[id] -
			u2->usignificand.bsignificand[id] - borrow;
		if (a != 0)
			diff = 1;	/* Suppose u1 > u2 until proven
					 * otherwise. */
		if (a < 0) {
			borrow = 1;
		} else {
			borrow = 0;
		}
		id++;
	}
	while (id < u1->usignificand.blength) {
		a = u1->usignificand.bsignificand[id] - borrow;
		if (a != 0)
			diff = 1;	/* Suppose u1 > u2 until proven
					 * otherwise. */
		if (a < 0) {
			borrow = 1;
		} else {
			borrow = 0;
		}
		id++;
	}
	while (id < u2->usignificand.blength) {
		a = -u2->usignificand.bsignificand[id] - borrow;
		if (a != 0)
			diff = 1;	/* Suppose u1 > u2 until proven
					 * otherwise. */
		if (a < 0) {
			borrow = 1;
		} else {
			borrow = 0;
		}
		id++;
	}
	if (borrow == 1) {	/* Sign reversal occurred - previous result
				 * invalid. */
		diff = -1;	/* u1 < u2 after all. */
	}
	return diff;
}

void
introduce_an_error(t)
	testrecord     *t;
{
	int		num_ulps;	/* How many ulps in error. */
	int		direction;	/* Add error if +1, subtract if -1. */

	/* Make the error large enough to guarantee we'll detect it.  If
	 * we happen to add in the error in the opposite direction of the
	 * true conversion error, it'll turn up as a rounding error; otherwise
	 * as an ULPS error.
	 */
	if ( curround == round_two )
		num_ulps = 4;
	else if ( curround == round_one )
		num_ulps = 2;
	else
		num_ulps = 1;

	if (random_fraction() >= 0.5)
		direction = 1;
	else
		direction = -1;

	if ( is_to_binary( curtest ) ) {
		binary_plus_nulps( FLOAT_ADDR t->b, direction * num_ulps,
				   FLOAT_ADDR t->b );

	} else {
		int		i, last_digit;

		last_digit = gstringpulp( t->d ) - t->d;

		for ( i = 0; i < num_ulps; ++i )
			(void) bump_decimalstring(
						direction, last_digit, t->d );
	}

	introduced_error = 1;
	++countintro;
}

int
one_test(t)
	testrecord     *t;

{

	/* Tests one decimal-to-binary or binary-to-decimal conversion
	 * and checks results.  Returns 1 if there was an ulps or
	 * rounding error, 0 otherwise.
	 */

	decimalstring   dr;

	/* Convert operand using code under test. */
	set_test_rounding( curround );

	if ( is_to_binary( curtest ) )
		testdb(t->d, FLOAT_ADDR (t->b), &(t->e));
	else
		testbd(FLOAT_ADDR (t->b), t->d, &cursig, &curtest, &(t->e));

	set_normal_rounding();

	++countconversions;

	introduced_error = 0;
	if ( introduce_errors && countconversions % introduce_errors == 0 )
		introduce_an_error( t );

	/* Unpack result to standard form. */
	binary_to_unpacked(FLOAT_ADDR(t->b), &(t->ub));

	/* Unpack operand to standard form. */
	decimalstring_to_unpacked(t->d, &(t->ud), dr);

	t->max_excess = -1;	/* We haven't figured the excess bits yet. */

	compare(t);

	return rounding_out_of_bounds( curround, t, 1 );
}

void
btod_monotests()
{
	/*
	 * Tests monotonicity for conversion from random binary to E/F/G
	 * format with n significant digits.
	 */

	FLOAT           newb;
	int             i, kmo, direction;
	testrecord     *old, *new;

	for (direction = 1; direction >= -1; direction -= 2) {
		kmo = 0;
		old = &center;
		for (i = 1; i <= nmono; i++) {	/* Increment by ulp. */
			binary_plus_nulps(FLOAT_ADDR(center.b),
					direction * i, FLOAT_ADDR newb);

			if (newb == old->b)
				/* Skip insignificant increments. */
				continue;

			kmo++;
			new = &(mo[kmo % 2]);
			new->dt = center.dt;
			new->index = center.index;
			new->offset = direction * i;
			copyfloat(FLOAT_ADDR newb, FLOAT_ADDR(new->b));

			if ( one_test(new) )
				return;

			if ( introduced_error )
				/* An introduced error may not lead to a
				 * monotonicity error until later, since
				 * it may have been a large step in what
				 * is the correct direction for monotonicity,
				 * but large enough so that the *next*
				 * test point will show a monotonicity
				 * error.  It's a pain to try to catch this,
				 * so we don't.
				 */
				break;

			if (mono_direction(&(old->ud),
					   &(new->ud)) == direction) {
				/* Monotonicity error. */
				countmono++;
				if ( ! testrecord_is_valid( &monoexample1 ) ||
				     ABS(monoexample1.ub.uexponent) >
				     ABS(old->ub.uexponent) ) {
					copy_testrecord(old, &monoexample1, 1);
					copy_testrecord(new, &monoexample2, 1);
				}

				if (verbosity == tedious) {
					(void) printf(
	    "At %d ulps MONOTONICITY ERROR: diff = dir %d\n", i, direction);
					(void) printf("MONO1");
					display_testrecord(&monoexample1);
					(void) printf("MONO2");
					display_testrecord(&monoexample2);
					dump_unpacked(&(old->ud));
					dump_unpacked(&(new->ud));
				}
			}

			old = new;
			if (old->b == 0)
				/* End this region if we cross zero. */
				break;
		}
	}
}

void
dtob_monotests()
{
	/*
	 * Tests monotonicity for conversion from random curtest format
	 * with field width n to binary.
	 */

	int             i, kmo, direction;
	testrecord     *old, *new;
	int             dulp;
	decimalstring   centerplusulps;
	int		monobound;

	monobound = (ulpsbound < nmono) ? ulpsbound : nmono;

	dulp = gstringpulp(center.d) - center.d;

	for (direction = 1; direction >= -1; direction -= 2) {
		kmo = 0;
		old = &center;
		(void) strcpy(centerplusulps, center.d);
		for (i = 1; i <= monobound; i++) {
			if ( bump_decimalstring( direction, dulp,
						 centerplusulps ) == 0 )
				/* Stop this test. */
				return;

			kmo++;
			new = &(mo[kmo % 2]);
			new->dt = center.dt;
			new->index = center.index;
			new->offset = direction * i;
			(void) strcpy(new->d, centerplusulps);

			if ( one_test(new) )
				return;

			if ( introduced_error )
				/* An introduced error may not lead to a
				 * monotonicity error until later, since
				 * it may have been a large step in what
				 * is the correct direction for monotonicity,
				 * but large enough so that the *next*
				 * test point will show a monotonicity
				 * error.  It's a pain to try to catch this,
				 * so we don't.
				 */
				break;

			if (new->ub.uclass != number)
				break;	/* begin a new direction */

			if (mono_direction(&(old->ub),
				    &(new->ub)) == direction) {
				/* Monotonicity error. */
				countmono++;
				if ( ! testrecord_is_valid( &monoexample1 ) ||
				     ABS(monoexample1.ub.uexponent) >
				     ABS(old->ub.uexponent) ) {
					copy_testrecord(old, &monoexample1, 1);
					copy_testrecord(new, &monoexample2, 1);
				}
				if (verbosity == tedious) {
					(void) printf(
			    "At %d ulps MONOTONICITY ERROR: diff = dir %d\n",
						i, direction);
					(void) printf("MONO1");
					display_testrecord(old);
					(void) printf("MONO2");
					display_testrecord(new);
					dump_unpacked(&(old->ub));
					dump_unpacked(&(new->ub));
				}
			}
			old = new;
		}
	}
}

void
finished_region()
{
	++countregions;

	if ( ! no_stress || countregions % 100 == 0 )
		printf( "." );

	if ( (no_stress && countregions % 5000 == 0) ||
	     (! no_stress && countregions % 50 == 0) )
		printf( " %lu regions, %lu conv's\n",
			countregions, countconversions);

	fflush( stdout );
}

void
test_regions()

{	/* do the current test for each region */

	int		num_regions;

	lastdata = startdata;

	initialize_testrecord(&ulpminpos);
	initialize_testrecord(&ulpmaxpos);
	initialize_testrecord(&ulpminneg);
	initialize_testrecord(&ulpmaxneg);

	initialize_testrecord(&excepexample);
	initialize_testrecord(&classexample);
	initialize_testrecord(&signexample);
	initialize_testrecord(&monoexample1);
	initialize_testrecord(&monoexample2);
	initialize_testrecord(&roundexample);
	initialize_testrecord(&whopperexample);

	countexcep = countclass = countsign = countmono = countround = 0;
	countintro = countconversions = countregions = 0;
	count_intro_caught = count_intro_missed = 0;

	if ( timelimit != 0.0 ) {
		timedout = 0;
		timestart = seconds();
	}

	if ( nregions )
		num_regions = nregions;

	else if ( no_stress )
		fatal( "nregions not set in test_regions()" );

	else {
		if ( is_to_binary( curtest ) )
			num_regions =
			max_bin_stress_exponent - min_bin_stress_exponent + 1;
		else
			num_regions =
			max_dec_stress_exponent - min_dec_stress_exponent + 1;
	}

	while ( countregions < num_regions ) {
		if ( timelimit != 0.0 ) {
			timestop = seconds();
			if (timestop > (timestart + timelimit)) {
				timedout = 1;
				break;
			}
		}

		/* Now do a random operand. */

		if ( is_to_binary( curtest ) )
			generate_data(curtest, (FLOAT *) 0, center.d);
		else
			generate_data(curtest, FLOAT_ADDR center.b, (char *) 0);

		if ( lastdata == stress && ! no_stress ) {
			generate_stress_tests();
			finished_region();

		} else {
			center.dt = lastdata;
			center.index = countregions;
			center.offset = 0;

			if ( one_test( &center ) == 0 &&
			     ! introduced_error && center.ulp.aulp >= 0 )
				if ( is_to_binary( curtest ) )
					dtob_monotests();
				else
					btod_monotests();

			if ( no_stress )
				finished_region();
		}
	}

	summary();
}

void
testloop()
{
	int             iround, itest;

	ulpsbound = 0;

	total_conversions = total_regions = 0;

	for (cursig = 3; cursig <= sigmin; cursig++)
		ulpsbound = 10 * ulpsbound + 9;

	for (cursig = sigmin; cursig <= sigmax; cursig++) {
		if (cursig > 1)
			ulpsbound = 10 * ulpsbound + 9;

		find_f_exponents();

		if ( test_enabled[(int) btof] ) {
			max_btof_dec_stress_exponent =
				MIN(max_dec_stress_exponent, max_10_f_exponent);
			min_btof_dec_stress_exponent =
				MAX(min_dec_stress_exponent, min_10_f_exponent);
			btof_dec_stress_exponent_stride =
				find_good_stride(
					max_btof_dec_stress_exponent -
					min_btof_dec_stress_exponent + 1 );
		}

		for (itest = 0; itest < (int) lasttesttype; itest++) {
		    curtest = (enum testtype) itest;

		    if (test_enabled[(int) curtest] == 1) {
			for (iround = 0;
			     iround < (int) lastroundingtype; iround++) {
			    if ( rounding_enabled[iround] ) {
				    curround = (enum roundingtype) iround;
				    test_regions();
				    total_conversions += countconversions;
				    total_regions += countregions;
			    }
			}
		    }
		}
	}
}
