/* Testbase routines for computing ULP values */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


void
binary_plus_nulps(pb, n, pbplus)
	FLOAT          *pb, *pbplus;
	int             n;

{
	/* Computes bplus = b + n ulps of b	 */

	floatsizeunion  u;
	FLOATSIZEUNIT   t;
	short unsigned  borrow, carry;
	int             i;

	copyfloat(pb, FLOAT_ADDR(u.f));
	if (n < 0) {
		borrow = -n;
		if (lsuindex == 0) {
			for (i = 0; i < floatsize; i++) {
				if (borrow != 0) {
					t = u.fsu[i] - borrow;
					if (u.fsu[i] >= t)
						borrow = 0;
					else
						borrow = 1;
					u.fsu[i] = t;
				}
			}
		} else {
			for (i = floatsize - 1; i >= 0; i--) {
				if (borrow != 0) {
					t = u.fsu[i] - borrow;
					if (u.fsu[i] >= t)
						borrow = 0;
					else
						borrow = 1;
					u.fsu[i] = t;
				}
			}
		}
	} else {
		carry = n;
		if (lsuindex == 0) {
			for (i = 0; i < floatsize; i++) {
				if (carry != 0) {
					t = u.fsu[i] + carry;
					if (u.fsu[i] <= t)
						carry = 0;
					else
						carry = 1;
					u.fsu[i] = t;
				}
			}
		} else {
			for (i = floatsize - 1; i >= 0; i--) {
				if (carry != 0) {
					t = u.fsu[i] + carry;
					if (u.fsu[i] <= t)
						carry = 0;
					else
						carry = 1;
					u.fsu[i] = t;
				}
			}
		}
	}

	canonify_bumped_binary( &u );
	copyfloat(FLOAT_ADDR(u.f), pbplus);
}

int
bump_decimalstring( direction, dulp, dstring )
	int		direction, dulp;
	decimalstring   dstring;
{
	/*
	 * Modifies dstring to reflect an increment or decrement
	 * (depnding on "direction") in the "dulp" digit.  Returns zero
	 * if we decremented the string to zero, otherwise returns non-zero.
	 */

	int             i;
	char		buf1, buf2;

	if (direction == 1) {
		/* Increment loop. */
		for (i = dulp; i >= 0; i--) {
			switch (dstring[i]) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
					dstring[i]++;
					return 1;

				case '9':
					dstring[i] = '0';
					break;
				}
		}

		/* Carried out of loop. */
		i = 0;
		while (dstring[i] != '0' && dstring[i] != '.')
			i++;

		/* Find first digit or point. */
		buf1 = '1';

		do {	/* Move string in place. */
			buf2 = dstring[i];
			dstring[i] = buf1;
			buf1 = buf2;
			i++;
		} while (dstring[i] != 0);

		dstring[i] = buf1;
		/* Terminate string. */
		dstring[i + 1] = 0;

		return 1;

	} else {	/* Decrement loop. */
		for (i = dulp; i >= 0; i--) {
			switch (dstring[i]) {
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					dstring[i]--;
					return 1;

				case '0':
					dstring[i] = '9';
					break;
				}
		}
		/* Carried out of loop - must've hit zero. */
		return 0;
	}
}

void
binary_halfulp(pb, pu)
	FLOAT          *pb, *pu;

{
	/* Computes u = 1/2 ulp of b	 */

	int             i;
	unpacked        u;

	binary_to_unpacked(pb, &u);

	for (i = 1; i < u.usignificand.blength; i++)
		u.usignificand.bsignificand[i] = 0;

	u.usignificand.bsignificand[0] = 1;	/* One ulp. */
	u.uexponent -= 1;	/* Now 1/2 ulp. */

	unpacked_to_binary(&u, pu);	/* *pu now has 1/2 ulp of *pb. */
}

void
decimal_plushalfulp(d, sig, du)
	decimalstring   d, du;
	enum fsign      sig;

{
	/*
	 * Computes du = d +- 0.5 ulp of d, where + or - is chosen according
	 * to sig.
	 */

	int             i, lastdigit, lastnonzero, dot, gap;

	/* Skip the uninteresting stuff.	 */
	i = 0;
	while ((('0' > d[i]) || (d[i] > '9')) && (d[i] != '.')) {
		du[i] = d[i];
		i++;
	}

	/* Find the digit string and dot.	 */
	lastdigit = -1;
	lastnonzero = -1;
	dot = DECIMALSTRING_LENGTH;

	while ((('0' <= d[i]) && (d[i] <= '9')) || (d[i] == '.')) {
		du[i] = d[i];
		if (d[i] == '.')
			dot = i;
		else {
			lastdigit = i;
			if (d[i] != '0')
				lastnonzero = i;
		}
		i++;
	}

	if ((lastdigit == -1) || ((lastnonzero == -1) && (sig == negative))) {
		/* Not a valid number or a zero to be decremented - just
		 * copy.
		 */
		(void) strcpy(du, d);

	} else {
		gap = i;	/* du[i] and du[i+1] aren't filled in yet. */

		/* Find the end of the string.	 */
		while (d[i] != 0) {
			du[i + 2] = d[i];
			i++;
		}
		du[i + 2] = 0;

		if (gap < (dot + 1)) {	/* String is dddddxyz with no dot. */
			du[gap] = '.';
			du[gap + 1] = '5';

		} else if (gap >= (dot + 1)) {	/* String is ddddd.xyz or
						 * ddd.ddd with dot.	 */
			du[gap] = '5';
			du[gap + 1] = '0';
		}

		if (sig == negative) {	/* Subtract half ulp must carry. */
			i = lastdigit;
			while ((du[i] == '0') || (du[i] == '.')) {
				if (du[i] == '0')
					du[i] = '9';
				i--;
			}
			du[i] -= 1;
		}
	}
}

int
binarylsb(pb)
	FLOAT          *pb;
{
	/*
	 * Returns least significant bit of *b to determine whether rounding
	 * ambiguous cases is correct.
	 */

	floatsizeunion *pu;

	pu = (floatsizeunion *) pb;
	return (pu->fsu[lsuindex] & 1);
}

char           *
gstringpulp(d)
	decimalstring   d;

{
	/* Returns pointer to ulp position of decimal string.	 */

	char           *pd;
	char           *lastdigit;
	pd = d;

	while ((*pd == ' ') || (*pd == '+') || (*pd == '-') || (*pd == '.')) {
		pd++;
	}			/* Copy irrelevant items. */
start:
	switch (*pd) {
	case '.':
		break;
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case '0':
		lastdigit = pd;
		break;
	default:
		goto done;	/* E e null etc. */
	}
	pd++;
	goto start;

done:

	return lastdigit;
}

void
ulpdiff(t)
	testrecord     *t;

{
	/*
	 * Computes (target - source) / ulp(target)
	 */

	EXPONENT        scaleBIG, scalex, scaley, scale2, scale10;
	WORK            diff, tdiff, huge, tiny;
	unpacked       *x, *y, scaledx, scaledy;
	bigint          diffint, halfulp;
	int             id, nhalves;
	int             r, q;

	/* Set up y to be target, x to be source */
	if ( is_to_binary( curtest ) ) {
		copy_unpacked(&(t->ub), &scaledy);
		copy_unpacked(&(t->ud), &scaledx);
	} else {
		copy_unpacked(&(t->ud), &scaledy);
		copy_unpacked(&(t->ub), &scaledx);
	}
	x = &scaledx;
	y = &scaledy;

	x->usignificand.bsign = x->usign;	/* Set up correct signs
						 * please! */
	y->usignificand.bsign = y->usign;

	/* Convert floating point to integers. */

	if (x->uexponent >= 0) {/* x form is definitely an integer. */
		bigint_times_power(&(x->usignificand), x->ubase, x->uexponent);
		scalex = 0;
	} else {		/* x form is probably a fraction. */
		/* scale y up instead of scaling x down */
		bigint_times_power(&(y->usignificand), x->ubase, -x->uexponent);
		scalex = -x->uexponent;
	}
	if (y->uexponent >= 0) {/* y form is definitely an integer. */
		bigint_times_power(&(y->usignificand), y->ubase, y->uexponent);
		scaley = y->uexponent;
	} else {		/* y form is probably a fraction. */
		/* scale x up instead of scaling y down */
		bigint_times_power(&(x->usignificand), y->ubase, -y->uexponent);
		scaley = 0;
	}

	(void) bigint_subtract(&(y->usignificand),
				&(x->usignificand), &diffint);
	/*
	 * diffint is y-x, so the sign of diffint is the sign of x or y if
	 * |y|>=|x|, and the opposite sign from x or y if |y|<|x| ; in other
	 * words, diffint has the true sign of the error
	 */

	/* Convert ulps to half-ulps. */
	bigint_times_power(&diffint, (unsigned) 2, (EXPONENT) 1);

	halfulp.bsign = y->usign;	/* Sign of ulp y == sign y	 */
	halfulp.blength = 1;	/* Compute 1 half-ulp of target. */
	halfulp.bsignificand[0] = 1;

	/* Compute one scaled half-ulp. */
	bigint_times_power(&halfulp, x->ubase, scalex);
	bigint_times_power(&halfulp, y->ubase, scaley);
	bigint_remainder(&diffint, &halfulp, &nhalves);

	/*
	 * Now the actual error y-x is 0.5 * nhalves * (1ulpy) + 0.5 *
	 * diffint
	 * 
	 * So the error in ulps of y is 0.5 * nhalves + 0.5 * diffint / (sign of
	 * y) (ulp of abs y)
	 */

	diff = diffint.bsignificand[0];

	for (id = 1; id < diffint.blength; ++id)
		diff = diffint.bsignificand[id] + diff / BIGBASE;

	if (diff != 0) {	/* Don't bother scaling zero! */
		scaleBIG = diffint.blength - 1;
		diff = diff / 2;

		if (diffint.bsign == negative)
			diff = -diff;	/* Sign of difference. */

		if (y->usign == negative)
			/* Adjust for sign of y to get ulps * of y. */
			diff = -diff;

		/*
		 * The scaled ulp difference is:
		 * 
		 * diff * BIGBASE ** scaleBIG
		 * 
		 * The actual difference in ulps is
		 * 
		 * diff * BIGBASE ** scaleBIG / xbase**scalex / ybase ** scaley
		 * 
		 * We want to compute actual ulps without gratuituous rounding
		 * errors nor underflow or overflow.  The first thing to do
		 * is convert the corrections into powers of 2 and 10, then
		 * apply them in a stable way.  So compute
		 * 
		 * diff * 2 ** (BIGBASE_BITS * scaleBIG - scale2) / 10**scale10
		 * 
		 * or letting scale10 = 3*q+r, 10**scale10 = 1000**q * 10**r =
		 * 1024**q * 8**r * (5/4)**r * (1000/1024)**q = 2**(10q+3r) *
		 * (5/4)**r * (1000/1024)**q
		 * 
		 * Hence we want
		 * 
		 * diff * 2**(BIGBASE_BITS * scaleBIG - scale2 - 10q -
		 * 3r)/(1.25**r * 0.9765625**q)
		 * 
		 */

		scale2 = BIGBASE_BITS * scaleBIG;
		scale10 = 0;
		if (x->ubase == 2) {
			scale2 -= scalex;
		}
		if (x->ubase == 10) {
			scale10 += scalex;
		}
		if (y->ubase == 2) {
			scale2 -= scaley;
		}
		if (y->ubase == 10) {
			scale10 += scaley;
		}
		q = scale10 / 3;
		r = scale10 % 3;
		scale2 -= 10 * q + 3 * r;

		/* Divide diff by 1.25**r * 0.9765625**q */

		while (q >= 1) {
			diff = diff / 0.9765625;
			q--;
		}
		while (r >= 1) {
			diff = diff / 1.25;
			r--;
		}

		/* Multiply diff by 2**scale2 */

		/* If |diff| gets bigger than "huge" then we just give up
		 * on computing diff more accurately.  We have to do this
		 * to avoid the risk of generating fatal overflows on
		 * machines such as Vaxes.  Similarly, if |diff| gets
		 * smaller than "tiny" then we give up to avoid fatal
		 * underflows.
		 */
		huge = 1e25;
		tiny = 1e-25;

		while (scale2 >= BIGBASE_BITS && ABS(diff) < huge) {
			diff = diff * BIGBASE;
			scale2 -= BIGBASE_BITS;
		}
		while (scale2 >= 1 && ABS(diff) < huge) {
			diff = 2.0 * diff;
			scale2--;
		};
		while (scale2 <= -BIGBASE_BITS && ABS(diff) > tiny) {
			tdiff = diff / BIGBASE;
			workdummy(&tdiff);
			if (tdiff == 0)
				break;	/* Watch out for scaling away to
					 * zero! */
			diff = tdiff;
			scale2 += BIGBASE_BITS;
		}
		while (scale2 <= -1 && ABS(diff) > tiny) {
			tdiff = diff / 2;
			workdummy(&tdiff);
			if (tdiff == 0)
				break;	/* Watch out for scaling away to
					 * zero! */
			diff = tdiff;
			scale2++;
		}

		/*
		 * Allow one extra rounding error in computing diff since the
		 * remainder may not fit in WORK precision if FLOAT precision
		 * is greater.
		 */

		if (diff > 0.25) {
			diff -= 0.50;
			nhalves++;
		}
		if (diff < -0.25) {
			diff += 0.50;
			nhalves--;
		}
	}
	t->ulp.halves = nhalves;
	t->ulp.extra = diff;
	diff = 0.5 * t->ulp.halves + t->ulp.extra;
	t->ulp.aulp = (diff >= 0) ? diff : -diff;
}

int
check_if_extreme_ulp( t )
	testrecord	*t;
{
	/* Returns true if the testrecord given in t is an extreme one
	 * and records it as the new example of the extreme; otherwise
	 * returns false.
	 */

	int		extreme_ulp = 0;

	if ( introduced_error )
		return 0;

	if ( ABS(t->ulp.halves) > 4 )
		copy_best_testrecord( &whopperexample, t );

	if (t->ub.usign == positive) {
		if (t->ulp.halves < ulpminpos.ulp.halves ||
		    (t->ulp.halves == ulpminpos.ulp.halves &&
		     t->ulp.extra < ulpminpos.ulp.extra)) {
			copy_testrecord(t, &ulpminpos, 0);
			extreme_ulp = 1;
		}
		if (t->ulp.halves > ulpmaxpos.ulp.halves ||
		    (t->ulp.halves == ulpmaxpos.ulp.halves &&
		     t->ulp.extra > ulpmaxpos.ulp.extra)) {
			copy_testrecord(t, &ulpmaxpos, 0);
			extreme_ulp = 1;
		}
	} else {
		if (t->ulp.halves < ulpminneg.ulp.halves ||
		    (t->ulp.halves == ulpminneg.ulp.halves &&
		     t->ulp.extra < ulpminneg.ulp.extra)) {
			copy_testrecord(t, &ulpminneg, 0);
			extreme_ulp = 1;
		}
		if (t->ulp.halves > ulpmaxneg.ulp.halves ||
		    (t->ulp.halves == ulpmaxneg.ulp.halves &&
		     t->ulp.extra > ulpmaxneg.ulp.extra)) {
			copy_testrecord(t, &ulpmaxneg, 0);
			extreme_ulp = 1;
		}
	}

	return extreme_ulp;
}
